/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 6, 2004
BacteriumApp.cpp: implementation of the BacteriumApp class.

							Application Structure
							---------------------
Initialize
	initSim()
Main processing loop
	Do actions for all polymers
		actSim()
			Grid::act()
				do all Polymer::act()
				do placeInGrid() for all Polymer
				do all Polymer::move()
	Do OpenGL redraw
		CallBackDisplayFunc()
			drawGrids()
				drawGrid1() for each of four layers

  Note:
  - both initSim() and actSim() are called by OpenGL CallBackTimerFunc()
**************************************************************************/

#include "BacteriumApp.h"
#include "ProteinFactory.h"
#include "BactEnvPdu.h"
#include <assert.h>

const int NumMcpTypes = 1; // 1=all TAR/aspartate; 4=all four MCP and ligands

// In client-server mode, the server controls most of the experiment timings
#define CLIENT_SERVER 1 // comment this out to turn off client-server mode

// Experiment Timings (in time steps)
// These specify time steps when these actions will occur.
#ifdef CLIENT_SERVER
// set high values so these events will never occur; server will do this
const int TimeToMaterializeGhosts = 999999; // release ligands into environment
const int TimeToMopupLigands      = 999999; // remove ligands from environment
const int TimeToCreateInventory   = 150000; // create inventory file
const int TimeToEndExperiment     = 999999; // pause the simulation
#else
// normal values when client determines these; no server
const int TimeToMaterializeGhosts =  50000; // release ligands into environment
const int TimeToMopupLigands      = 180000; // remove ligands from environment
const int TimeToCreateInventory   = 150000; // create inventory file
const int TimeToEndExperiment     = 200000; // pause the simulation
#endif

/*
// Example of equivalent char and int strings for the same protein:
char templateStr[] = "SDWRY*hhhh**EDGR**G**KGR**";

int templateStr[] = {
	SerMN, AspMN, TrpMN, ArgMN, TyrMN,
	WildCardNoGlyMN,
	HydrophobicMN, HydrophobicMN, HydrophobicMN, HydrophobicMN,
	WildCardNoGlyMN, WildCardNoGlyMN,
	GluMN,
	AspMN, GlyMN, ArgMN,
	WildCardNoGlyMN, WildCardNoGlyMN, GlyMN, WildCardNoGlyMN, WildCardNoGlyMN,
	LysMN, GlyMN, ArgMN, WildCardNoGlyMN, WildCardNoGlyMN,
	NullMN
};
*/

// Logic Recognizer strings (cytoplasm proteins)
// Full set of 16 truth table (tt) recognizer strings
// Example (AND):
//     Input1 Input2 | Result
//     ------ ------ | ------
//       0      0    |   0
//       0      1    |   0
//       1      0    |   0
//       1      1    |   1
// w or W (Trp) must match water in the grid
// q or Q (Gln) delimiter; signifies stop trying to match
char tt0000RecogStr[] = "DDDDD";       // 0 0 0 0
char tt1000RecogStr[] = "DRKRD";       // 1 0 0 0
char tt0100RecogStr[] = "wDRKRE";      // 0 1 0 0
char tt1100RecogStr[] = "wDRKR";       // 1 1 0 0
char tt0010RecogStr[] = "wERKRD";      // 0 0 1 0
char tt1010RecogStr[] = "DRKRw";       // 1 0 1 0
char tt0110RecogStr[] = "ERKRD";       // 0 1 1 0 XOR  (not linearly separable)
char tt1110RecogStr[] = "DRKR";        // 1 1 1 0
char tt0001RecogStr[] = "ERKRE";       // 0 0 0 1 AND
char tt1001RecogStr[] = "DRKRDqERKRE"; // 1 0 0 1 (not linearly separable)
char tt0101RecogStr[] = "ERKRw";       // 0 1 0 1
char tt1101RecogStr[] = "ERKREqRKRDw"; // 1 1 0 1 (input 1 zero, or both ones)
char tt0011RecogStr[] = "wERKR";       // 0 0 1 1
char tt1011RecogStr[] = "DRKRDqRKREw"; // 1 0 1 1 (input 1 one, or both zeros)
char tt0111RecogStr[] = "ERKR";        // 0 1 1 1 OR
char tt1111RecogStr[] = "RKR";         // 1 1 1 1
char *andRecogStr = tt0001RecogStr; // AND Recognizer
char *orRecogStr  = tt0111RecogStr; // OR
char *xorRecogStr = tt0110RecogStr; // XOR

// ligand small molecules as int and char strings
//int ligandA[] = {SmD_MN, SmW_MN, SmR_MN, SmY_MN, NullMN}; // DWRY
//int ligandB[] = {SmS_MN, SmD_MN, SmW_MN, SmR_MN, NullMN}; // SDWR
char ligandA[] = "DWRY";
char ligandB[] = "SDWR";

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BacteriumApp::BacteriumApp( GlutMaster* gm, Probabilities* pr, Statistics *st )
{
	glutMaster = gm;
	prob = pr;
	stats = st;
	gridSizeX = 120;
	gridSizeY = 120;
	logicType = NullLT;
	trialType = 0;
	unitSize = 6; // default: 5   useful range: 1 to 8

	// these values are read in from file if appNum == 3, 5 or 6
	numLigands = 0;
	numAxp = 0;
	numMeth = 0;
	numMeth0 = 0;
	numCheY = 4;
	numCheB = 0;
	numCheR = 0;
	numCheZ = 0;
	ligOffsetX = 0;
	ligOffsetY = 0;

	timeStepCounter = 0;
	itsPerTimeStep = 1;
	step = false;
	appNum = 0;
	pause = false;
	firstTime = true;
	outQ = NULL;
	inQ = NULL;
	pauseDisplay = false;
}

BacteriumApp::~BacteriumApp()
{

}

////////////////////////////////////////////////
// CallBackTimerFunc
// Called by OpenGL at start of each time step.
////////////////////////////////////////////////
void BacteriumApp::CallBackTimerFunc(int value)
{
	// Initialize
	if (firstTime) {
		firstTime = false;
		switch (appNum) {
		case 1:
			initSim_1(value); // Logic application; 00 01 10 11
			break;
		case 2:
			initSim_2(value); // test
			break;
		case 3:
			initSim_3(value); // Bacterial Chemotaxis
			break;
		case 4:
			initSim_4(value); // test
			break;
		case 5:
			initSim_5(value); // Bacterial Chemotaxis (test of self-assembly)
			break;
		case 6:
			initSim_6(value); // Bacterial Chemotaxis (larger bacterium)
			break;
		default:
			break;
		}
	}
	// Regular action each time step
	else
		actSim( value );
}

////////////////////////////////
// initSim_1
// Initialize Logic application
////////////////////////////////
void BacteriumApp::initSim_1( int tVal )
{
	int i;

	if (Msg::debugM) printf( "initSim_1()\n" );

	// Lipid bilayer
	lipBilayer = new LipidBilayer( grid, 1, 18, Rot0, FlipNone, MembraneLY, NULL, false, 87, 105);
	grid->addPolymer( lipBilayer );
	lipBilayer = new LipidBilayer( grid, 23, 46, Rot0, FlipNone, MembraneLY, NULL, false, 40, 40);
	if (Msg::infoM) printf( "\nLipid Bilayer\n" );
	lipBilayer->printShape();
	grid->addPolymer( lipBilayer );

	// Ligands
	int numLigands = 1; // 10
	char *ligStr1, *ligStr2;
	int y1, y2;
	switch (trialType) {
	case 0: ligStr1 = ligandA; ligStr2 = ligandA; y1=90; y2=90; break; // 00
	case 1: ligStr1 = ligandA; ligStr2 = ligandB; y1=90; y2=91; break; // 01
	case 2: ligStr1 = ligandB; ligStr2 = ligandA; y1=91; y2=90; break; // 10
	case 3: ligStr1 = ligandB; ligStr2 = ligandB; y1=91; y2=91; break; // 11
	default: printf( "ERROR: initSim %d\n", trialType ); break;
	}
	if (Msg::infoM) printf( "\nLigands (LIG)\n" );
	for (i = 0; i < numLigands; i++) {

		// Ligand (LIGA) inhibition 0 -1
		lig = new SmallMolecule(grid, 38, y1, Rot90, FlipNone, SmallMolLY,
			NULL, false, ligStr1);
		lig->printShape();
		grid->addPolymer( lig );

		// Ligand (LIGB) activation 1 +1
		lig = new SmallMolecule(grid, 42, y2, Rot90, FlipNone, SmallMolLY,
			NULL, false, ligStr2);
		lig->printShape();
		grid->addPolymer( lig );
	}

	// Transmembrane protein (TMPA) with 3 binding sites
	char *tmpProt = "SDWRYhhhh**EDGR**G**KGR**";
	prot = new Protein(grid, 34, 83, Rot90, FlipNone, Protein1LY,
		lipBilayer, false, tmpProt, prob, "TmpA");
	prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
	prot->setRuleVal( 0, 1, InhibitPistonBindingRUL );
	if (Msg::infoM) printf( "\nTransmembrane Protein (TMP)\n" );
	prot->printString();
	prot->printShape();
	grid->addPolymer( prot );

	// Another instance of transmembrane protein (TMPB) with 3 binding sites
	prot = new Protein(grid, 38, 83, Rot90, FlipNone, Protein1LY,
		lipBilayer, false, tmpProt, prob, "TmpB");
	prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
	prot->setRuleVal( 0, 1, InhibitPistonBindingRUL );
	if (Msg::infoM) printf( "\nTransmembrane Protein (TMP)\n" );
	prot->printString();
	prot->printShape();
	grid->addPolymer( prot );

	// Cytoplasm protein (CP) with 2 binding sites.
	// Select the string for the test specified in the config file.
	int numCp = 100;
	char* ttRecogStr;
	switch( logicType) {
	case _0000_LT:
		ttRecogStr = tt0000RecogStr;
		break;
	case _1000_LT:
		ttRecogStr = tt1000RecogStr;
		break;
	case _0100_LT:
		ttRecogStr = tt0100RecogStr;
		break;
	case _1100_LT:
		ttRecogStr = tt1100RecogStr;
		break;
	case _0010_LT:
		ttRecogStr = tt0010RecogStr;
		break;
	case _1010_LT:
		ttRecogStr = tt1010RecogStr;
		break;
	case _0110_LT: // XorLT
		ttRecogStr = tt0110RecogStr; // xorRecogStr
		break;
	case _1110_LT:
		ttRecogStr = tt1110RecogStr;
		break;
	case _0001_LT: // AndLT
		ttRecogStr = tt0001RecogStr; // andRecogStr
		break;
	case _1001_LT:
		ttRecogStr = tt1001RecogStr;
		break;
	case _0101_LT:
		ttRecogStr = tt0101RecogStr;
		break;
	case _1101_LT:
		ttRecogStr = tt1101RecogStr;
		break;
	case _0011_LT:
		ttRecogStr = tt0011RecogStr;
		break;
	case _1011_LT:
		ttRecogStr = tt1011RecogStr;
		break;
	case _0111_LT: // OrLT
		ttRecogStr = tt0111RecogStr; // orRecogStr
		break;
	case _1111_LT:
		ttRecogStr = tt1111RecogStr;
		break;
/*
	case AndLT:
		prot->setString( andRecogStr );
		break;
	case OrLT:
		prot->setString( orRecogStr );
		break;
	case XorLT:
		prot->setString( xorRecogStr );
		break;
*/
	default:
		break;
	}
/*
	// Show all 16 recognizers; used to get screen print
	prot = new Protein(grid, 26, 82, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0000RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 80, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0001RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 78, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0010RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 76, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0011RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 74, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0100RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 72, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0101RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 70, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0110RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 68, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt0111RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 66, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1000RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 64, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1001RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 62, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1010RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 60, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1011RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 58, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1100RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 56, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1101RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 54, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1110RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
	prot = new Protein(grid, 26, 52, Rot0, FlipNone, Protein2LY,
		lipBilayer, false, tt1111RecogStr, prob, "Recognizer");
	grid->addPolymer( prot );
*/
	for (i = 0; i < numCp; i++) {
		prot = new Protein(grid, 38, 74, Rot0, FlipNone, Protein2LY,
			lipBilayer, false, ttRecogStr, prob, "Recognizer");
		prot->setRuleVal( 0, 0, ProteinProteinBindingRUL );
		// these three logic types have a second binding domain
		if ((logicType == _1001_LT)
			|| (logicType == _1011_LT)
			|| (logicType == _1101_LT)) {
			prot->setRuleVal( 0, 6, ProteinProteinBindingRUL );
		}
		grid->addPolymer( prot );
	}
	prot->printString();
	prot->printShape();

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_1()

///////////////////////////////
// initSim_2
// Initialize test application
///////////////////////////////
void BacteriumApp::initSim_2( int tVal )
{
	int i;

	if (Msg::debugM) printf( "initSim_2()\n" );

	// Lipid bilayer
	LipidBilayer *lipBil1 = new LipidBilayer( grid, 3, 23, Rot0, FlipNone, MembraneLY, NULL, false, 80, 105);
	grid->addPolymer( lipBil1 );
	LipidBilayer *lipBil2 = new LipidBilayer( grid, 10, 53, Rot0, FlipNone, MembraneLY, NULL, false, 40, 30);
	grid->addPolymer( lipBil2 );
	LipidBilayer *lipBil3 = new LipidBilayer( grid, 60, 33, Rot0, FlipNone, MembraneLY, NULL, false, 30, 43);
	grid->addPolymer( lipBil3 );

	// Ligands
	int numLigands = 4000; // stress test using a large number of ligands
	for (i = 0; i < numLigands; i++) {
		lig = new SmallMolecule(grid, 45, 64, Rot0, FlipNone, SmallMolLY,
			lipBil1, false, "STYH");
		grid->addPolymer( lig );
	}

	// Hydrophobic transmembrane proteins (TMP) free to move
	for (i = 0; i < 20; i++) {
		prot = new Protein(grid, 20, 70, Rot0, FlipNone, Protein1LY,
			lipBil2, false, "DhhhhDGE", prob);
		grid->addPolymer( prot );
	}

	// Hydrophobic transmembrane proteins (TMP) free to move in outer membrane
	for (i = 0; i < 20; i++) {
		prot = new Protein(grid, 45, 55, Rot0, FlipNone, Protein1LY,
			lipBil1, false, "BRKRDhhhhDGG*****", prob);
		grid->addPolymer( prot );
	}

	// Random proteins inside a membrane
	for (i = 0; i < 20; i++) { // 50
		prot = new Protein(grid, 70, 40, Rot0, FlipNone, Protein1LY,
			lipBil3, false, "###G#####G######G###", prob);
		grid->addPolymer( prot );
		prot = new Protein(grid, 75, 45, Rot0, FlipNone, Protein1LY,
			lipBil3, false, "#######G###", prob);
		grid->addPolymer( prot );
	}

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_2()

/////////////////////////////////////////////////////
// initSim_3
// Initialize Bacterial Chemotaxis (small bacterium)
/////////////////////////////////////////////////////
void BacteriumApp::initSim_3( int tVal )
{
	int i;
	int scX, scY; // sensory complex coordinates
	int nCheY = numCheY; // to allow having CheY = 0

	if (Msg::debugM) printf( "initSim_3()\n" );

	// use ProteinFactory as a convenience to reduce the amount of code
	ProteinFactory *pf = new ProteinFactory( this );

	// Lipid bilayer
	lipBilayer = new LipidBilayer( grid, 1, 18, Rot0, FlipNone, MembraneLY, NULL, false, 87, 105);
	grid->addPolymer( lipBilayer );
	lipBilayer = new LipidBilayer( grid, 10, 21, Rot0, FlipNone, MembraneLY, NULL, false, 60, 40);
	lipBilayer->setProcessRunsTumbles( true, outQ, windowID );
	grid->addPolymer( lipBilayer );
	lipBilayer->printShape();

	// Ligands
	addLigands();

	// ATP
	SmallMolecule *atp;
	for (i = 0; i < numAxp; i++) {
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AtpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AdpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AmpSM, "Axp");
		grid->addPolymer( atp );
	}
	atp->printShape();

	// Methanol
	SmallMolecule *meth;
	for (i = 0; i < numMeth; i++) {
		meth = new SmallMolecule(grid, 35, 65, Rot0, FlipNone, SmallMolLY,
			NULL, false, MethSM, "Meth");
		grid->addPolymer( meth );
	}
	meth->printShape();
	for (i = 0; i < numMeth0; i++) {
		meth = new SmallMolecule(grid, 35, 65, Rot0, FlipNone, SmallMolLY,
			NULL, false, Meth0SM, "Meth0");
		grid->addPolymer( meth );
	}
	meth->printShape();

	// Motors
	// Motor complex; all as one protein for now
	prot = pf->create( Motor_PT, 42, 58, "MotorR1" );
	prot = pf->create( Motor_PT, 42, 50, "MotorR2" );
	prot = pf->create( Motor_PT, 42, 42, "MotorR3" );
	prot = pf->create( Motor_PT, 5, 58, "MotorL1" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, 5, 50, "MotorL2" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, 5, 42, "MotorL3" ); prot->setRot(Rot180);

	// TAR sensory complex
	scX = 10; scY = 77;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TSR sensory complex
	scX = 18;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tsr_PT, scX, scY, "TSR" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TAP sensory complex
	scX = 26;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tap_PT, scX, scY, "TAP" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TRG sensory complex
	scX = 34;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Trg_PT, scX, scY, "TRG" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// free floating CheY - should bind with CheA once other CheY dissociates
	for (i = 0; i < nCheY; i++) {
		prot = pf->create( CheY_PT, 17, 60, "CheY" );
	}

	// free floating CheB
	for (i = 0; i < numCheB; i++) {
		prot = pf->create( CheB_PT, 21, 55, "CheB" );
	}

	// free floating CheR
	for (i = 0; i < numCheR; i++) {
		prot = pf->create( CheR_PT, 24, 50, "CheR" );
	}

	// free floating CheZ
	for (i = 0; i < numCheZ; i++) {
		prot = pf->create( CheZ_PT, 28, 45, "CheZ" );
	}

	// test phosphorylated protein
	//prot = pf->create( PhosTest_PT, 20, 60, "PhosTest" );

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_3()

////////////////////////////////////////////////////
// initSim_4
// Initialize test of protein matching for proteins
// in all rotated and flipped configurations.
////////////////////////////////////////////////////
void BacteriumApp::initSim_4( int tVal )
{
	int i;

	if (Msg::debugM) printf( "initSim_4()\n" );

	char* protStr = "AVFVAGGE"; // all are hydrophobic

	// Lipid bilayer
	lipBilayer = new LipidBilayer( grid, 0, 25, Rot0, FlipNone, MembraneLY, NULL, NULL, 78, 108);
	grid->addPolymer( lipBilayer );

	// Set up attached unmoving proteins.
	// Normally proteins should not be pre-specified as attached,
	// and should never be attached to nothing as they are here, but this is just a test.
	// Rot0
	prot = new Protein(grid, 20, 80, Rot0, FlipNone, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 30, 80, Rot0, FlipVert, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 40, 80, Rot0, FlipHorz, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 50, 80, Rot0, FlipBoth, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	// Rot90
	prot = new Protein(grid, 20, 70, Rot90, FlipNone, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 30, 70, Rot90, FlipVert, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 40, 70, Rot90, FlipHorz, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 50, 70, Rot90, FlipBoth, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	// Rot180
	prot = new Protein(grid, 20, 60, Rot180, FlipNone, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 30, 60, Rot180, FlipVert, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 40, 60, Rot180, FlipHorz, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 50, 60, Rot180, FlipBoth, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	// Rot270
	prot = new Protein(grid, 20, 50, Rot270, FlipNone, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 30, 50, Rot270, FlipVert, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 40, 50, Rot270, FlipHorz, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );
	prot = new Protein(grid, 50, 50, Rot270, FlipBoth, Protein1LY,
		lipBilayer, true, protStr, prob);
	grid->addPolymer( prot );

	// test
	prot = new Protein(grid, 15, 72, Rot0, FlipNone, Protein1LY,
		lipBilayer, false, "DhhhhDGE", prob);
	grid->addPolymer( prot );

	// set up unattached moving proteins
	for (i = 0; i < 100; i++) {
		prot = new Protein(grid, 35, 65, Rot0, FlipNone, Protein2LY,
			lipBilayer, false, "AVFVA", prob);
		prot->setRuleVal( 0, 0, ProteinProteinBindingRUL );
		grid->addPolymer( prot );
	}

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_4()

////////////////////////////////////////////////////////////////////////////////////
// initSim_5
// Initialize Bacterial Chemotaxis (test of self-assembly)
// Some of the proteins should bind with the membrane and then bind with each other.
// All proteins are initially placed in the center of the cytoplasm.
////////////////////////////////////////////////////////////////////////////////////
void BacteriumApp::initSim_5( int tVal )
{
	int i;
	int cntrX = 24; int cntrY = 50;

	if (Msg::debugM) printf( "initSim_5()\n" );

	ProteinFactory *pf = new ProteinFactory( this );

	// Lipid bilayer
	lipBilayer = new LipidBilayer( grid, 1, 18, Rot0, FlipNone, MembraneLY, NULL, false, 87, 105);
	grid->addPolymer( lipBilayer );
	lipBilayer = new LipidBilayer( grid, 10, 21, Rot0, FlipNone, MembraneLY, NULL, false, 60, 40);
	lipBilayer->setProcessRunsTumbles( true, outQ, windowID );
	grid->addPolymer( lipBilayer );


	// Aspartate
	for (i = 0; i < numLigands; i++) {
		lig = new SmallMolecule(grid, 23, 90, Rot0, FlipNone, SmallMolLY,
			NULL, false, "DDDD", "Aspartate");
		grid->addPolymer( lig );
	}
	// Serine
	for (i = 0; i < numLigands; i++) {
		lig = new SmallMolecule(grid, 28, 90, Rot90, FlipNone, SmallMolLY,
			NULL, false, "SSSS", "Serine");
		grid->addPolymer( lig );
	}
	// Dipeptide
	for (i = 0; i < numLigands; i++) {
		lig = new SmallMolecule(grid, 33, 91, Rot180, FlipNone, SmallMolLY,
			NULL, false, "WHRD", "Dipeptide");
		grid->addPolymer( lig );
	}
	// Ribose & galactose
	for (i = 0; i < numLigands; i++) {
		lig = new SmallMolecule(grid, 38, 92, Rot270, FlipNone, SmallMolLY,
			NULL, false, "NSHW", "Ribose_Galactose");
		grid->addPolymer( lig );
	}

	// ATP
	SmallMolecule *atp;
	for (i = 0; i < 20; i++) {
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AtpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AdpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 35, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AmpSM, "Axp");
		grid->addPolymer( atp );
	}

	// Methanol
	SmallMolecule *meth;
	for (i = 0; i < 20; i++) {
		meth = new SmallMolecule(grid, 25, 38, Rot0, FlipNone, SmallMolLY,
			NULL, false, MethSM, "Meth");
		grid->addPolymer( meth );
	}

	// Motors
	// Motor complex; all as one protein for now
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorR1" );
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorR2" );
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorR3" );
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorL1" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorL2" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, cntrX, cntrY, "MotorL3" ); prot->setRot(Rot180);

	// TAR sensory complex
	prot = pf->create( Tar_PT,  cntrX, cntrY, "TAR" );
	prot = pf->create( CheW_PT, cntrX, cntrY, "CheW" );
	prot = pf->create( CheA_PT, cntrX, cntrY, "CheA" );
	prot = pf->create( CheY_PT, cntrX, cntrY, "CheY" );

	// TSR sensory complex
	prot = pf->create( Tsr_PT, cntrX, cntrY, "TSR" );
	prot = pf->create( CheW_PT, cntrX, cntrY, "CheW" );
	prot = pf->create( CheA_PT, cntrX, cntrY, "CheA" );
	prot = pf->create( CheY_PT, cntrX, cntrY, "CheY" );

	// TAP sensory complex
	prot = pf->create( Tap_PT, cntrX, cntrY, "TAP" );
	prot = pf->create( CheW_PT, cntrX, cntrY, "CheW" );
	prot = pf->create( CheA_PT, cntrX, cntrY, "CheA" );
	prot = pf->create( CheY_PT, cntrX, cntrY, "CheY" );

	// TRG sensory complex
	prot = pf->create( Trg_PT, cntrX, cntrY, "TRG" );
	prot = pf->create( CheW_PT, cntrX, cntrY, "CheW" );
	prot = pf->create( CheA_PT, cntrX, cntrY, "CheA" );
	prot = pf->create( CheY_PT, cntrX, cntrY, "CheY" );

	// free floating CheY - should bind with CheA once other CheY dissociates
	for (i = 0; i < numCheY-4; i++) { // -4 because four are already created above
		prot = pf->create( CheY_PT, cntrX, cntrY, "CheY" );
	}

	// free floating CheB
	for (i = 0; i < numCheY; i++) {
		prot = pf->create( CheB_PT, cntrX, cntrY, "CheB" );
	}
	prot->printShape();

	// free floating CheR
	for (i = 0; i < numCheY; i++) {
		prot = pf->create( CheR_PT, cntrX, cntrY, "CheR" );
	}

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_5()

///////////////////////////////////////////
// initSim_6
// Bacterial Chemotaxis - larger bacterium
///////////////////////////////////////////
void BacteriumApp::initSim_6( int tVal )
{
	int i;
	int scX, scY; // sensory complex coordinates
	int nCheY = numCheY; // to allow having CheY = 0

	if (Msg::debugM) printf( "initSim_6()\n" );

	ProteinFactory *pf = new ProteinFactory( this );

	// Lipid bilayer
	lipBilayer = new LipidBilayer( grid, 1, 18, Rot0, FlipNone, MembraneLY, NULL, false, 87, 105);
	grid->addPolymer( lipBilayer );
	lipBilayer = new LipidBilayer( grid, 10, 31, Rot0, FlipNone, MembraneLY, NULL, false, 60, 85);
	lipBilayer->setProcessRunsTumbles( true, outQ, windowID );
	grid->addPolymer( lipBilayer );
	lipBilayer->printShape();

	// Ligands
	addLigands();

	// ATP
	SmallMolecule *atp;
	for (i = 0; i < numAxp; i++) {
		atp = new SmallMolecule(grid, 60, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AtpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 60, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AdpSM, "Axp");
		grid->addPolymer( atp );
		atp = new SmallMolecule(grid, 60, 60, Rot0, FlipNone, SmallMolLY,
			NULL, false, AmpSM, "Axp");
		grid->addPolymer( atp );
	}
	atp->printShape();

	// Methanol
	SmallMolecule *meth;
	for (i = 0; i < numMeth; i++) {
		meth = new SmallMolecule(grid, 55, 65, Rot0, FlipNone, SmallMolLY,
			NULL, false, MethSM, "Meth");
		grid->addPolymer( meth );
	}
	meth->printShape();
	for (i = 0; i < numMeth0; i++) {
		meth = new SmallMolecule(grid, 55, 65, Rot0, FlipNone, SmallMolLY,
			NULL, false, Meth0SM, "Meth0");
		grid->addPolymer( meth );
	}
	meth->printShape();

	// Motors
	// Motor complex; all as one protein for now
	prot = pf->create( Motor_PT, 87, 68, "MotorR1" );
	prot = pf->create( Motor_PT, 87, 60, "MotorR2" );
	prot = pf->create( Motor_PT, 87, 52, "MotorR3" );
	prot = pf->create( Motor_PT, 5, 68, "MotorL1" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, 5, 60, "MotorL2" ); prot->setRot(Rot180);
	prot = pf->create( Motor_PT, 5, 52, "MotorL3" ); prot->setRot(Rot180);

	// TAR sensory complex
	scX = 10; scY = 87;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	//prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TSR sensory complex
	scX = 16;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tsr_PT, scX, scY, "TSR" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	//prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TAP sensory complex
	scX = 22;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Tap_PT, scX, scY, "TAP" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	//prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// TRG sensory complex
	scX = 28;
	if (NumMcpTypes == 4) // all four MCP types
		prot = pf->create( Trg_PT, scX, scY, "TRG" );
	else if (NumMcpTypes == 1) // only TAR and aspartate
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
	//prot = pf->create( CheW_PT, scX+7, scY-4, "CheW" );
	prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
	if (nCheY-- > 0)
		prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );

	// more TAR sensory complexes along top
	for (i = 0; i < 9; i++) {
		scX += 6;
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
		prot = pf->create( CheA_PT, scX+3, scY-5, "CheA" );
		if (nCheY-- > 0)
			prot = pf->create( CheY_PT, scX+5, scY-6, "CheY" );
	}

	// even more TAR sensory complexes along bottom
	scY = 30; scX = 4;
	for (i = 0; i < 13; i++) {
		scX += 6;
		prot = pf->create( Tar_PT,  scX, scY, "TAR" );
		prot->setRot( Rot270 );
		prot = pf->create( CheA_PT, scX+3, scY+7, "CheA" );
		prot->setRot( Rot180 );
		if (nCheY-- > 0)
			prot = pf->create( CheY_PT, scX+3, scY+8, "CheY" );
			prot->setRot( Rot180 );
	}

	// free floating CheY - should bind with CheA once other CheY dissociates
	for (i = 0; i < nCheY; i++) {
		prot = pf->create( CheY_PT, 47, 65, "CheY" );
	}

	// free floating CheB
	for (i = 0; i < numCheB; i++) {
		prot = pf->create( CheB_PT, 51, 60, "CheB" );
	}

	// free floating CheR
	for (i = 0; i < numCheR; i++) {
		prot = pf->create( CheR_PT, 54, 55, "CheR" );
	}

	// free floating CheZ
	for (i = 0; i < numCheZ; i++) {
		prot = pf->create( CheZ_PT, 58, 50, "CheZ" );
	}

	grid->setNoMoreRules();
	grid->act();
	glutPostRedisplay();
} // end initSim_6()

//////////////////////////////////////////////////////////////////////////////////////
// actSim
// Regular action performed every time step.
// Do processing only if the simulation is not paused, or if it's been single-stepped.
//////////////////////////////////////////////////////////////////////////////////////
void BacteriumApp::actSim( int tVal )
{
	if (!pause && !globalPause) {
		for (int i = 0; i < itsPerTimeStep; i++) {
			grid->act();
			timeStepCounter++;
			timeExperiment();
		}
		glutPostRedisplay();
	}
	else if (step || globalStep) {
		grid->act();
		timeStepCounter++;
		timeExperiment();
		glutPostRedisplay();
		step = false;
	}
}

////////////////////////////////////////////////////////////////////////
// timeExperiment
// Do various experiment actions that need to be done at certain times.
////////////////////////////////////////////////////////////////////////
void BacteriumApp::timeExperiment( void)
{
	// release ligands all at once
	if (timeStepCounter == TimeToMaterializeGhosts)
		grid->materializeGhosts(numLigands * LigandsMultiplier * 4); // 4 types

	// or release ligands gradually
	//if (timeStepCounter >= TimeToMaterializeGhosts) {
	//	if ((timeStepCounter % 1000) == 0)
	//		grid->materializeGhosts(numLigands * 0.2);
	//}

	// will only receive a PDU if there is a separate server app
	rxPdu();

	if (timeStepCounter == TimeToCreateInventory)
		grid->inventory( windowID, mn );

	if (timeStepCounter == TimeToMopupLigands)
		grid->mopUpLigands();

	if (timeStepCounter == TimeToEndExperiment) {
		pause = true;
		globalPause = true;
	}
}

/////////////////////////////////////////////////////
// rxPdu
// optionally receive a PDU from the EnvironmentApp.
// a PDU is a message transmitted through a queue.
/////////////////////////////////////////////////////
void BacteriumApp::rxPdu( void )
{
	BactEnvPdu *bep;
	while ((bep = (BactEnvPdu*)inQ->dequeue()) != Q_EMPTY) {
		if (bep->pduType == GradientBPT) {
			if (bep->pduVal >= 1)
				grid->materializeGhosts( bep->pduVal );
			else if (bep->pduVal <= -1)
				grid->mopUpLigands( bep->pduVal * (-1) );
		}
		else { // error
			printf( "ERROR: invalid pduType %d\n", bep->pduType );
		}
		delete bep; // all done with it so delete what server created
	}
}

/////////////////////////////////
// drawGrids
// Draw Grids in graphics window.
/////////////////////////////////
void BacteriumApp::drawGrids()
{
	//if (Msg::debugM) printf( "drawGrids\n" );
	gint* monoId;

	// draw membrane layer
	monoId = grid->getLayerMono( MembraneLY );
	drawGrid1( monoId, 0.0f, 0.0f, 0.0f );
	monoId = grid->getLayerMono( SmallMolLY );
	drawGrid1( monoId, 0.0f, 0.0f, 0.0f );
	monoId = grid->getLayerMono( Protein1LY );
	drawGrid1( monoId, 0.0f, 0.0f, 0.0f );
	monoId = grid->getLayerMono( Protein2LY );
	drawGrid1( monoId, 0.0f, 0.5f, 0.0f );

	if (unitSize < 4) { // may want to take different actions with different sizes
		//printf( "Time Step: %d", timeStepCounter);
		//return;
	}

	// Display informative text in window.
	char *tStr;
    glColor3ub(0, 0, 0);
	Text::setfont("helvetica", 18);
	if (unitSize == 2)
		Text::setfont("helvetica", 18);
	if (appNum == 1) {
		Text::drawstr(0, 150, "Logic: ");
		switch (logicType) {
		case _0000_LT: tStr = "0000"; break;
		case _1000_LT: tStr = "1000"; break;
		case _0100_LT: tStr = "0100"; break;
		case _1100_LT: tStr = "1100"; break;
		case _0010_LT: tStr = "0010"; break;
		case _1010_LT: tStr = "1010"; break;
		case _0110_LT: tStr = "XOR";  break;
		case _1110_LT: tStr = "1110"; break;
		case _0001_LT: tStr = "AND";  break;
		case _1001_LT: tStr = "1001"; break;
		case _0101_LT: tStr = "0101"; break;
		case _1101_LT: tStr = "1101"; break;
		case _0011_LT: tStr = "0011"; break;
		case _1011_LT: tStr = "1011"; break;
		case _0111_LT: tStr = "OR";   break;
		case _1111_LT: tStr = "1111"; break;
		default: tStr = "???"; break;
		}
		Text::drawstr(90, 150, tStr);
		Text::drawstr(130, 150, ", Inputs: ");
		switch (trialType) {
		case 0: tStr = "0 0,"; break;
		case 1: tStr = "0 1,"; break;
		case 2: tStr = "1 0,"; break;
		case 3: tStr = "1 1,"; break;
		default: tStr = "? ?,"; break;
		}
		Text::drawstr(200, 150, tStr);
	}
	if (unitSize == 2) {
		Text::drawstr(0, 18, "TimeStep: ");
		Text::drawstr(88, 18, "%d", timeStepCounter);
		Text::drawstr(168, 18, "WinID: %d", windowID);
	}
	else {
		Text::drawstr(230, 150, "Time Step: ");
		Text::drawstr(330, 150, "%d", timeStepCounter);
		Text::drawstr(430, 150, "WinID: %d", windowID);
	}
}

////////////////////////////
// drawGrid1
// Draw grid for one layer.
////////////////////////////
void BacteriumApp::drawGrid1( gint* monoId, float rOff, float gOff, float bOff )
{
	int row, col;
	for (row = 0; row < gridSizeY; row++) {
		for (col = 0; col < gridSizeX; col++) {
			mt = mn->get( monoId[row*gridSizeX+col] ); // get monomer type

			// the background is already water so it doesn't need to be displayed
			if (mt->id != WaterMN) {
				// specially set to display CP proteins in orangish-brown; not very elegant
				glColor3f( mt->colorRed+rOff, mt->colorGreen+gOff*mt->colorRed, mt->colorBlue+bOff );
				glRectd(unitSize*col, unitSize*row+unitSize, unitSize*col+unitSize, unitSize*row);
			}
		}
	}
}

////////////////////////////////////
// CallBackDisplayFunc
// Display everything in the grids.
// OpenGL (GlutMaster) callback function.
////////////////////////////////////
void BacteriumApp::CallBackDisplayFunc( void )
{
	if (pauseDisplay) return;
    // Make background colour Water
	mt = mn->get( WaterMN );
    glClearColor( mt->colorRed, mt->colorGreen, mt->colorBlue, 0 );
    glClear( GL_COLOR_BUFFER_BIT );
	drawGrids();
    glutSwapBuffers();
}

///////////////////////////////////////
// CallBackReshapeFunc
// Redraw the window if it's reshaped.
// OpenGL (GlutMaster) callback function.
///////////////////////////////////////
void BacteriumApp::CallBackReshapeFunc(int w, int h)
{
	if (Msg::debugM) printf( "reshape\n" );
	if (Msg::infoM) printf( "width: %d  height: %d\n", w, h );
	glOrtho(0, w, 0, h, -1, 1);
	if (unitSize > 3) // optional translation, rotation, and scale
		glTranslatef( 50.0f, -100.0f, 0.0f ); // reposition everything
	//glRotatef( 10.0f, 0.0f, 0.0f, 1.0f );
	//glScalef( 0.5f, 0.5f, 1.0f );

	/* suggested in redbook.pdf p.33
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluOrtho2D (0.0, (GLdouble) w, 0.0, (GLdouble) h);
	*/
}

//////////////////////////////////////////
// CallBackKeyboardFunc
// Handle keyboard input.
// OpenGL (GlutMaster) callback function.
//////////////////////////////////////////
void BacteriumApp::CallBackKeyboardFunc(unsigned char key, int x, int y)
{
    if (key==27) // ESC key to quit
		exit(0);
	else if (key == 'p')  { // pause
		if (pause) pause = false;
		else pause = true;
	}
	else if (key == 'P') { // global pause
		if (globalPause) globalPause = false;
		else globalPause = true;
	}
	else if (key == 's') // step
		step = true;
	else if (key == 'S') // globalStep
		globalStep = true;
	else if ((key == 'd') || (key == 'D')) { // inventory
		if (pauseDisplay) pauseDisplay = false;
		else pauseDisplay = true;
	}
	else if ((key == 'i') || (key == 'I')) // inventory
		grid->inventory( windowID, mn );
	else if (key == 'l') // ligands
		grid->materializeGhosts(numLigands);
	else if (key == 'L') // all ligands
		grid->materializeGhosts(numLigands * LigandsMultiplier * 4); // 4 types
	else if ((key == 'm') || (key == 'M')) // mop up ligands
		grid->mopUpLigands();
	else if ((key == 'h') || (key == 'H')) { // help
		printf( "Bacterium Application (client):\n" );
		printf( "  p pause this window/app only\n" );
		printf( "  P pause all windows/apps\n" );
		printf( "  s step this paused window/app only\n" );
		printf( "  S step all paused windows/apps\n" );
		printf( "  d pause display only\n" );
		printf( "  i create inventory of all polymers\n" );
		printf( "  l create numLigands ligands\n" );
		printf( "  L create all ligands\n" );
		printf( "  m mop up all ligands\n" );
		printf( "  h help\n" );
		printf( "  Esc exit from all windows/apps\n" );
	}
}

///////////////////////////////////////////////////////////////
// CallBackMouseFunc
// Handle mouse motion actions (not used for anything useful).
// OpenGL (GlutMaster) callback function.
///////////////////////////////////////////////////////////////
void BacteriumApp::CallBackMouseFunc(int button, int state, int x, int y)
{
	if (state == GLUT_UP)
		if (Msg::appM) printf("x:%d y:%d\n", x, winSizeY - y);
}

/////////////////////////////////////
// glMain
// OpenGL and other initializations.
/////////////////////////////////////
void BacteriumApp::glMain( int argc, char *argv[] )
{
	grid = new Grid( gridSizeY, gridSizeX, maxPolymers, prob, stats );
	mn = new Monomers();
	winSizeX = grid->getNumCols() * unitSize;
	winSizeY = grid->getNumRows() * unitSize;
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowSize( winSizeX, winSizeY );
	glutMaster->CallGlutCreateWindow("PBNN", this);
}

////////////////////////
// tRan
// test random numbers.
////////////////////////
void BacteriumApp::tRan( void )
{
	Rndm::testRandom(1, SizeOfActionEnumACT, 1000);
}

///////////////////////////////
// tGrid2shape
// test grid2shape() function.
///////////////////////////////
void BacteriumApp::tGrid2shape( void )
{
	// A different transmembrane protein (TMP) free to move
	prot = new Protein(grid, 12, 10, Rot270, FlipVert, Protein1LY,
		lipBilayer, NULL, "ACDEFGHIKLMGPQRSTGVWYACGDEFHI", prob);
	prot->printShape();
	//grid->addPolymer( prot );
	int cont, bind, gPos;
	int gY;
	int gX;
	int sRow = 0;
	int sCol = 0;
	prot->shape2grid( GyGxS2G, sRow, sCol, &gY, &gX );
	printf( "BEFORE: sRow=%d sCol=%d gY=%d gX=%d\n", sRow, sCol, gY, gX );
	prot->shape2grid( GposContS2G, sRow, sCol, &gPos, &cont );
	printf( "BEFORE: sRow=%d sCol=%d gPos=%d cont=%d\n", sRow, sCol, gPos, cont );
	prot->grid2shape( 0, gY, gX, &sRow, &sCol, &cont, &bind );
	printf( "AFTER: sRow=%d sCol=%d cont=%d bind=%d\n", sRow, sCol, cont, bind );
}

///////////////////
// initialize
///////////////////
void BacteriumApp::initialize( char *configFileName )
{
	// Read values from configuration file
	readConfigFromFile( configFileName );
	stats->cnt[Ligands_ST] = numLigands; // number from config file
}

////////////////////////////////////////////
// readConfigFromFile
// Read configuration values from a file.
////////////////////////////////////////////
void BacteriumApp::readConfigFromFile( char *fileName )
{
	FILE *fp;
	char *mode = "r";
	const int sizeBuff = 100;
	char buffer[sizeBuff];
	// variables that values from file get temporarily read into
	char ttStr[10];
	char ltStr[10];
	int intVal;
	int msgVal1, msgVal2, msgVal3;
	bool showParms = false;

	if ((fp = fopen(fileName, mode)) == NULL) {
		if (Msg::errorM) printf( "Can't open file: %s in %s mode\n", fileName, mode);
		exit(0);
	}
	else {
		// printf("Opened file: %s in %s mode\n", fileName, mode);
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d", &intVal ) != 1) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (intVal == 1)
				showParms = true;
		}
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d %d %d", &msgVal1, &msgVal2, &msgVal3 ) != 3) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			// appM
			if (msgVal1 == 0)
				Msg::appM = false;
			else
				Msg::appM = true;
			// infoM
			if (msgVal2 == 0)
				Msg::infoM = false;
			else
				Msg::infoM = true;
			// errorM
			if (msgVal3 == 0)
				Msg::errorM = false;
			else
				Msg::errorM = true;
			if (showParms) {
				printf( "Msg::appM   : %d\n", Msg::appM );
				printf( "Msg::infoM  : %d\n", Msg::infoM );
				printf( "Msg::errorM : %d\n", Msg::errorM );
			}
		}
		// gridSizeX, gridSizeY
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d %d", &gridSizeX, &gridSizeY ) != 2) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "gridSizeX : %d\n", gridSizeX );
				printf( "gridSizeY : %d\n", gridSizeY );
			}
		}
		// unitSize
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d", &unitSize ) != 1) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "unitSize : %d\n", unitSize );
			}
		}
		// itsPerTimeStep
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d", &itsPerTimeStep ) != 1) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "itsPerTimeStep : %d\n", itsPerTimeStep );
			}
		}
		// appNum
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d", &appNum ) != 1) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "appNum : %d\n", appNum );
			}
		}
		// config values specific to appNum 1
		if (appNum == 1) { // logic types AND OR XOR etc.
			// rescan the same buffer
			if (sscanf( buffer, "%d %s %s", &appNum, ttStr, ltStr ) != 3) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "inputs    : %s\n", ttStr );
				printf( "logicType : %s\n", ltStr );
			}
			getTrialTypeFromFile( ttStr );
			getLogicTypeFromFile( ltStr );
		}
		// config values specific to appNum 3, 5, 6
		if ((appNum == 3) || (appNum == 5) || (appNum == 6)) {
			// rescan the same buffer
			if (sscanf( buffer, "%d %d %d %d %d %d %d %d %d %d %d", &appNum,
				&numLigands, &numAxp, &numMeth, &numMeth0,
				&numCheY, &numCheB, &numCheR, &numCheZ,
				&ligOffsetX, &ligOffsetY ) != 11) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (showParms) {
				printf( "numLigands : %d\n", numLigands );
				printf( "numAxp : %d\n", numAxp );
				printf( "numMeth : %d\n", numMeth );
				printf( "numMeth0 : %d\n", numMeth0 );
				printf( "numCheY : %d\n", numCheY );
				printf( "numCheB : %d\n", numCheB );
				printf( "numCheR : %d\n", numCheR );
				printf( "numCheZ : %d\n", numCheZ );
				printf( "ligOffsetX : %d\n", ligOffsetX );
				printf( "ligOffsetY : %d\n", ligOffsetY );
			}
		}
		// optionally read in probabilities
		if (fgets(buffer, sizeBuff, fp) != NULL) {
			if (sscanf( buffer, "%d", &intVal ) != 1) {
				if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
				exit(0);
			}
			if (intVal == 1) { // 1=true; read probabilities from file
				if (sscanf( buffer, "%d %d %d %d %d %d %d %d %d", &intVal,
					&prob->pr[ProbLigandUnbindPRI],
					&prob->pr[ProbCheAAutoPhosPRI],
					&prob->pr[ProbCheYAutoDephosPRI],
					&prob->pr[ProbCheBAutoDephosPRI],
					&prob->pr[ProbMotorCwPRI],
					&prob->pr[ProbPistonBindToLigandPRI],
					&prob->pr[ProbCheBdemethInactivePRI],
					&prob->pr[ProbCheRmethActivePRI] ) != 9) {
					if (Msg::errorM) printf("Application: error reading from %s\n", fileName);
					exit(0);
				}
				if (showParms) {
					printf( "ReadProbabiliesFromFile : %d\n", intVal );
					printf( "ProbLigandUnbindPRI : %d\n", prob->pr[ProbLigandUnbindPRI] );
					printf( "ProbCheAAutoPhosPRI : %d\n", prob->pr[ProbCheAAutoPhosPRI] );
					printf( "ProbCheYAutoDephosPRI : %d\n", prob->pr[ProbCheYAutoDephosPRI] );
					printf( "ProbCheBAutoDephosPRI : %d\n", prob->pr[ProbCheBAutoDephosPRI] );
					printf( "ProbMotorCwPRI : %d\n", prob->pr[ProbMotorCwPRI] );
					printf( "ProbPistonBindToLigandPRI : %d\n", prob->pr[ProbPistonBindToLigandPRI] );
					printf( "ProbCheBdemethInactivePRI : %d\n", prob->pr[ProbCheBdemethInactivePRI] );
					printf( "ProbCheRmethActivePRI : %d\n", prob->pr[ProbCheRmethActivePRI] );
				}
			}
			else if (showParms) {
				printf( "ReadProbabiliesFromFile : %d\n", intVal );
			}
		}
		fclose(fp);
	}
}

/////////////////////////////
// getTrialTypeFromFile
// Get trial type from file.
/////////////////////////////
void BacteriumApp::getTrialTypeFromFile( char* ttStr )
{
	trialType = 0;
	if (!strcmp(ttStr, "00"))
		trialType = 0;
	else if (!strcmp(ttStr, "01"))
		trialType = 1;
	else if (!strcmp(ttStr, "10"))
		trialType = 2;
	else if (!strcmp(ttStr, "11"))
		trialType = 3;
}

/////////////////////////////
// getLogicTypeFromFile
// Get logic type from file.
/////////////////////////////
void BacteriumApp::getLogicTypeFromFile( char* ltStr )
{
	if (!strcmp(ltStr, "AND"))
		logicType = AndLT;
	else if (!strcmp(ltStr, "OR"))
		logicType = OrLT;
	else if (!strcmp(ltStr, "XOR"))
		logicType = XorLT;
	else if (!strcmp(ltStr, "0000"))
		logicType = _0000_LT;
	else if (!strcmp(ltStr, "1000"))
		logicType = _1000_LT;
	else if (!strcmp(ltStr, "0100"))
		logicType = _0100_LT;
	else if (!strcmp(ltStr, "1100"))
		logicType = _1100_LT;
	else if (!strcmp(ltStr, "0010"))
		logicType = _0010_LT;
	else if (!strcmp(ltStr, "1010"))
		logicType = _1010_LT;
	else if (!strcmp(ltStr, "0110"))
		logicType = _0110_LT;
	else if (!strcmp(ltStr, "1110"))
		logicType = _1110_LT;
	else if (!strcmp(ltStr, "0001"))
		logicType = _0001_LT;
	else if (!strcmp(ltStr, "1001"))
		logicType = _1001_LT;
	else if (!strcmp(ltStr, "0101"))
		logicType = _0101_LT;
	else if (!strcmp(ltStr, "1101"))
		logicType = _1101_LT;
	else if (!strcmp(ltStr, "0011"))
		logicType = _0011_LT;
	else if (!strcmp(ltStr, "1011"))
		logicType = _1011_LT;
	else if (!strcmp(ltStr, "0111"))
		logicType = _0111_LT;
	else if (!strcmp(ltStr, "1111"))
		logicType = _1111_LT;
	else {
		if (Msg::errorM) printf( "Invalid logic type %s\n", ltStr );
		exit(0);
	}
}

////////////////////////////////////////////////////////////
// setQ
// Set an inQ or outQ for optional use with EnvironmentApp.
////////////////////////////////////////////////////////////
void BacteriumApp::setQ( Queue *q, int direction )
{
	if (direction == Q_OUT)
		outQ = q;
	else
		inQ = q;
}

////////////////////////////////////////////////////////////////
// addLigands
// add a quantity of ligands = numLigands * 5
// all are initially "ghosts"
// be able to release numLigands each time user presses 'L' key
////////////////////////////////////////////////////////////////
void BacteriumApp::addLigands( void )
{
	int i;
	int ligX;
	int ligY;
	int nLigs = numLigands * LigandsMultiplier;

	ligX = 25 + ligOffsetX;
	ligY = 92 + ligOffsetY;

	if (appNum == 6) {
		ligX = 50 + ligOffsetX;
		for (i = 0; i < nLigs*4; i++) { // do half above and half below
			if ((i % 2) == 0)
				ligY = 99 + ligOffsetY;
			else
				ligY = 24 + ligOffsetY;
			lig = new SmallMolecule(grid, ligX, ligY-1, Rot0, FlipNone, SmallMolLY,
				NULL, false, "DDDD", "Aspartate");
			lig->setGhost( true );
			grid->addPolymer( lig );
		}
		return;
	}

	if (NumMcpTypes == 1) { // only TAR and aspartate
		// Aspartate
		for (i = 0; i < nLigs*4; i++) {
			lig = new SmallMolecule(grid, ligX, ligY-1, Rot0, FlipNone, SmallMolLY,
				NULL, false, "DDDD", "Aspartate");
			lig->setGhost( true );
			grid->addPolymer( lig );
		}
		return;
	}

	// NumMcpTypes == 4
	// Aspartate
	for (i = 0; i < nLigs; i++) {
		lig = new SmallMolecule(grid, ligX, ligY-1, Rot0, FlipNone, SmallMolLY,
			NULL, false, "DDDD", "Aspartate");
		lig->setGhost( true );
		grid->addPolymer( lig );
	}
	// Serine
	for (i = 0; i < nLigs; i++) {
		lig = new SmallMolecule(grid, ligX, ligY, Rot0, FlipNone, SmallMolLY,
			NULL, false, "SSSS", "Serine");
		lig->setGhost( true );
		grid->addPolymer( lig );
	}
	// Dipeptide
	for (i = 0; i < nLigs; i++) {
		lig = new SmallMolecule(grid, ligX, ligY+1, Rot0, FlipNone, SmallMolLY,
			NULL, false, "WHRD", "Dipeptide");
		lig->setGhost( true );
		grid->addPolymer( lig );
	}
	// Ribose & galactose
	for (i = 0; i < nLigs; i++) {
		lig = new SmallMolecule(grid, ligX, ligY+2, Rot0, FlipNone, SmallMolLY,
			NULL, false, "NSHW", "Ribose_Galactose");
		lig->setGhost( true );
		grid->addPolymer( lig );
	}
}
