/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 5, 2004
Grid.cpp: implementation of the Grid class.

**************************************************************************/

#include "Grid.h"
#include "Monomer.h"
#include "Msg.h"
#include "Rndm.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// don't use this constructor
Grid::Grid() { }

// use this constructor
Grid::Grid( int nRows, int nCols, int maxPoly, Probabilities* pr, Statistics *st )
{
	prob = pr;
	stats = st;

	numRows = nRows;
	numCols = nCols;
	gridSize = numRows * numCols;

	maxPolymers = maxPoly;
	polymer = new Polymer*[maxPolymers];
	numCurrentPolymers = 0;
	doActions = false;

	LMembraneMonoId = new gint[gridSize];
	LMembranePolyId = new gint[gridSize];
	LProtein1MonoId = new gint[gridSize];
	LProtein1PolyId = new gint[gridSize];
	LProtein2MonoId = new gint[gridSize];
	LProtein2PolyId = new gint[gridSize];
	LSmallMolMonoId = new gint[gridSize];
	LSmallMolPolyId = new gint[gridSize];
	initGrids();
}

Grid::~Grid()
{

}

//////////////////
// Get functions
//////////////////
int Grid::getNumRows( void ) {return numRows;}
int Grid::getNumCols( void ) {return numCols;}
int Grid::getGridSize( void ) {return gridSize;}
int Grid::getNumCurrentPolymers( void ) {return numCurrentPolymers;}

//////////////////////////////////////
// getLayerMono
// Get a specific monomer grid layer.
//////////////////////////////////////
gint* Grid::getLayerMono( int gridLayer )
{
	gint *monoId = NULL;

	switch (gridLayer) {
	case MembraneLY:
		monoId = LMembraneMonoId;
		break;
	case Protein1LY:
		monoId = LProtein1MonoId;
		break;
	case Protein2LY:
		monoId = LProtein2MonoId;
		break;
	case SmallMolLY:
		monoId = LSmallMolMonoId;
		break;
	default:
		break;
	}
	return monoId;
}

//////////////////////////////////////
// getLayerPoly
// Get a specific polymer grid layer.
//////////////////////////////////////
gint* Grid::getLayerPoly( int gridLayer )
{
	gint *polyId = NULL;

	switch (gridLayer) {
	case MembraneLY:
		polyId = LMembranePolyId;
		break;
	case Protein1LY:
		polyId = LProtein1PolyId;
		break;
	case Protein2LY:
		polyId = LProtein2PolyId;
		break;
	case SmallMolLY:
		polyId = LSmallMolPolyId;
		break;
	default:
		break;
	}
	return polyId;
}

//////////////////////////////////////////////////
// addPolymer
// Add a polymer to the list of current polymers.
//////////////////////////////////////////////////
void Grid::addPolymer( Polymer* pol )
{
	if (numCurrentPolymers < maxPolymers) {
		pol->setID( numCurrentPolymers );
		polymer[numCurrentPolymers++] = pol;
	}
	else {
		if (Msg::errorM)
			printf( "ERROR: Grid::addPolymer() maxPolymers exceeded %d\n", numCurrentPolymers );
	}
}

///////////////////////////////////////////////
// getPolymer
// Get polymer corresponding to a specific ID.
///////////////////////////////////////////////
Polymer* Grid::getPolymer( int ix )
{
	if ((ix < numCurrentPolymers) && (ix >= 0))
		return polymer[ix];
	else {
		if (Msg::errorM)
			printf( "ERROR: Grid::getPolymer() invalid polymer ID %d\n", ix );
		return NULL;
	}
}

////////////////////////////////////////////////////////
// act
// Do various actions for every polymer each time step.
////////////////////////////////////////////////////////
void Grid::act( void )
{
	int i;
	Polymer* pol;
	initGrids();
	// 1. Do any non-movement actions
	//    - binding to polymers already in the grid
	//    - conformational changes such as TMP piston; etc.
	// 2. Place all polymers in grids
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			if (pol->getGhost()) continue;
			if (doActions) {
				pol->act(); // do regular action
				handleEntropy( pol ); // do things that are caused by the environment
				adjustBindings( pol ); // check if bindings still match
			}
			placeInGrid( pol, i ); // put polymer in the grid
			if (stats) {
				pol->doStats(); // stats
			}
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::act() NULL polymer2\n" );
		}
	}
	if (stats) { // not a particularly good way to check for something
		stats->output(); // output statistics
	}
	if (!doActions) {
		doActions = true;
		return;
	}
	// 3. Do polymer random movements
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			if (pol->getGhost()) continue;
			pol->move(); // move
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::act() NULL polymer3\n" );
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// handleEntropy
// Done in Grid class rather than in Polymer classes, because entropy
// or degradation is something imposed from the outside world on the polymer,
// and is not something the polymer originates itself.
// Use probabilities to decide:
//   - if bindings should be broken
//   - if polymers should be broken up into monomers, or otherwise degraded
//////////////////////////////////////////////////////////////////////////////
void Grid::handleEntropy( Polymer* pol1 )
{
	// if gridLayer == SmallMolLY (then this is a small molecule)
	//   if probabilistically determine that should unbind
	//     move in a random direction (right, left, up, down, flip, rotate)
	//       pol->move(), which will disallow a random move if overlaps membrane
	//     set change to true, so adjustBindings can do its thing
	int rNum;
	if (pol1->getGhost()) return;
	if (pol1->getGridLayer() == SmallMolLY) { // this is a small molecule
		if (pol1->getAttached()) { // if it's attached to something (a protein)
			rNum = Rndm::getRandomInt( ProbabilityMax-1 );
			if ((rNum < prob->pr[ProbLigandUnbindPRI])
				|| (pol1->getShapeVal(0,0) == SmA_MN)) {  // this one is supposed to be mopped up
				pol1->move(); // attempt a random move
				pol1->setChanged( true );
			}
		}
	}
}

///////////////////////////////////////////////////////////
// adjustBindings
// Unbind bindings if the this polymer is no longer bound.
///////////////////////////////////////////////////////////
void Grid::adjustBindings( Polymer* pol1 )
{
	if (!(pol1->getAttached())) return;
	if (!(pol1->getChanged())) return;
	if (pol1->getGhost()) return;
	pol1->setChanged( false );
	int row, col;
	int opID; // other polymer ID
	int nCols = pol1->getNumCols();
	int nRows = pol1->getNumRows();
	Polymer* pol2 = NULL;
	int mon1, mon2;
	int gRow, gCol;
	int row2, col2, bind2;

	for (row = 0; row < nRows; row++) {
		for (col = 0; col < nCols; col++) {
			opID = pol1->getBindingVal(row, col);
			if (opID != NullBNG) {
				pol2 = getPolymer( opID );
				mon1 = pol1->getShapeVal(row, col);
				// get mon2
				pol1->shape2grid( GyGxS2G, row, col, &gRow, &gCol );
				pol2->grid2shape( 0, gRow, gCol, &row2, &col2, &mon2, &bind2 );
				if (!matchM_M( pol1, pol2, (enum MonomerTypeEnum)mon1, (enum MonomerTypeEnum)mon2 )) { // no longer matches
					pol1->unbind( opID );
					pol2->unbind( pol1->getID() );
					if (pol1->getShapeVal(0,0) == SmA_MN) { // needs to complete being mopped up
						pol1->setShapeVal(0,0,SmD_MN);
						pol1->setGhost( true );
					}
				}
			}
		}
	}
}

//////////////////////////////
// matchM_M
// Try to match two monomers.
//////////////////////////////
bool Grid::matchM_M( Polymer* pol1, Polymer* pol2, MonomerTypeEnum mon1, MonomerTypeEnum mon2 )
{
	bool rc;
	if ((mon1 == mon2) && pol1->isAminoAcid( mon1 )) // Protein - Protein
		rc = true;
	else if (pol1->isHydrophobic( mon1 ) && pol2->isLipid( mon2 )) // Protein - Membrane
		rc = true;
	else if (pol2->isHydrophobic( mon2 ) && pol1->isLipid( mon1 )) // Protein - Membrane
		rc = true;
	else if (mon1+(SmA_MN-AlaMN) == mon2) // Protein - SM
		rc = true;
	else if (mon2+(SmA_MN-AlaMN) == mon1) // Protein - SM
		rc = true;
	else
		rc = false;
	return rc;
}

//////////////////////////////////////////////////////////////////////////////
// setNoMoreRules
// To increase performance, place a special rule in the 2D shape of a protein
// indicating that this is the last rule and that the rest of the shape need
// not be searched for additional rules.
//////////////////////////////////////////////////////////////////////////////
void Grid::setNoMoreRules( void )
{
	int i;
	Polymer* pol;
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			pol->setNoMoreRules();
		}
	}
}

/////////////////////////////////////////////////////////
// placeInGrid
// Place polymer in the grid corresponding to its layer.
/////////////////////////////////////////////////////////
void Grid::placeInGrid( Polymer* pol, int ix )
{
	int layer = pol->getGridLayer();
	int nRows = pol->getNumRows(); // size of input
	int nCols = pol->getNumCols();
	int row, col;
	gint *monoId = getLayerMono( layer );
	gint *polyId = getLayerPoly( layer );

	int gridPos = 0;
	int contents = 0;
	for (row = 0; row < nRows; row++) {
		for (col = 0; col < nCols; col++) {
			pol->shape2grid( GposContS2G, row, col, &gridPos, &contents );
			if (contents != WaterMN) { // water is just filler
				monoId[gridPos] = contents;
				polyId[gridPos] = ix;
			}
		}
	}
}

/////////////////////////////////////////////////////
// initGrids
// Initialize all layer grids before writing to them.
//////////////////////////////////////////////////////
void Grid::initGrids( void )
{
	int i;
	for (i = 0; i < gridSize; i++) {
		LMembraneMonoId[i] = WaterMN;
		LMembranePolyId[i] = NullPolyID;
		LProtein1MonoId[i] = WaterMN;
		LProtein1PolyId[i] = NullPolyID;
		LProtein2MonoId[i] = WaterMN;
		LProtein2PolyId[i] = NullPolyID;
		LSmallMolMonoId[i] = WaterMN;
		LSmallMolPolyId[i] = NullPolyID;
	}
}

/////////////////////////////////////////////////
// inventory
// Write an inventory of all polymers to a file.
/////////////////////////////////////////////////
void Grid::inventory( int clientNum, Monomers *mn )
{
	int i;
	Polymer* pol;
	FILE *fp;
	char fileName[100];
	char *modeW = "w"; // Open to Write

	// Get current time
	time_t tnow;
	struct tm *tmnow;
	time( &tnow );
	tmnow = localtime( &tnow );

	// include date and time as part of file name
	sprintf( fileName, "..\\..\\Inventory\\%c%02d%02d%02d_%02d%02d%02d_inv%d.csv\0",
		statsSeries, tmnow->tm_mon + 1, tmnow->tm_mday, tmnow->tm_year - 100,
		tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec, clientNum );
	// open file
	if ((fp = fopen(fileName, modeW)) == NULL) {
		printf( "Can't open file: %s in %s mode\n", fileName, modeW);
		return;
	}
	// write header information
	fprintf( fp, "EASY MSc Dissertation - Protein-Based Neural Networks\n" );
	fprintf( fp, "Inventory of Polymers in C++ Simulation\n" );
	fprintf( fp, "Ken Webb\n" );
	fprintf( fp, "%s\n", fileName );
	fprintf( fp, "=(rule) |(binding) #(both rule and binding) .(no more rules)\n" );

	// write data for all current polymers
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			pol->inventory( fp, mn );
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::inventory() NULL polymer\n" );
		}
	}
	// close file
	fclose( fp );
}

///////////////////////////////////////////////////////////////
// materializeGhosts
// Make ligands appear and become an active part of the world.
// These are "ghosts" until they become activated.
///////////////////////////////////////////////////////////////
void Grid::materializeGhosts( int num2Materialize )
{
	int i;
	Polymer* pol;
	int mCount = 0;
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			if (pol->getGhost()) {
				pol->setGhost( false );
				mCount++ ;
				if (mCount >= num2Materialize)
					break;
			}
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::materializeGhosts() NULL polymer\n" );
		}
	}
	if (Msg::infoM)
		printf( "materialized %d out of %d\n", mCount, num2Materialize );
}

///////////////////////////////////////
// mopUpLigands
// Return all ligands to ghost status.
///////////////////////////////////////
void Grid::mopUpLigands( void )
{
	int i;
	Polymer* pol;
	char* nm;
	int mCount = 0;
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			nm = pol->getName();
			// This is a very slow way to check if something is a ligand,
			// and is also a maintenance issue.
			if ((strcmp(nm, "Aspartate") == 0)
				|| (strcmp(nm, "Serine") == 0)
				|| (strcmp(nm, "Dipeptide") == 0)
				|| (strcmp(nm, "Ribose_Galactose") == 0)) {
				if ((!pol->getAttached()) && (!pol->getGhost())) {
					pol->setGhost( true );
					mCount++ ;
				}
				// not a particularly good way to do this, but it works
				// make it a ghost later as soon as it unbinds
				else if ((pol->getAttached()) // disable re-binding
						&& (pol->getShapeVal(0,0) != SmA_MN)) {
					pol->setShapeVal(0,0,SmA_MN);
					mCount++ ;
				}
			}
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::mopUpLigands() NULL polymer\n" );
		}
	}
	if (Msg::infoM)
		printf( "mopped up %d\n", mCount );
}

/////////////////////////////////////////////////////
// mopUpLigands
// Return a given number of ligands to ghost status.
/////////////////////////////////////////////////////
void Grid::mopUpLigands( int num2MopUp )
{
	int i;
	Polymer* pol;
	char* nm;
	int mCount = 0;
	for (i = 0; i < numCurrentPolymers; i++) {
		pol = getPolymer( i );
		if (pol != NULL) {
			nm = pol->getName();
			if ((strcmp(nm, "Aspartate") == 0)
				|| (strcmp(nm, "Serine") == 0)
				|| (strcmp(nm, "Dipeptide") == 0)
				|| (strcmp(nm, "Ribose_Galactose") == 0)) {
				if ((!pol->getAttached()) && (!pol->getGhost())) {
					pol->setGhost( true );
					mCount++;
					if (mCount >= num2MopUp)
						break;
				}
				else if ((pol->getAttached()) // disable re-binding
					&& (pol->getShapeVal(0,0) != SmA_MN)) {
					pol->setShapeVal(0,0,SmA_MN);
					mCount++;
					if (mCount >= num2MopUp)
						break;
				}
			}
		}
		else {
			if (Msg::errorM)
				printf( "ERROR: Grid::mopUpLigand1() NULL polymer\n" );
		}
	}
	if (Msg::infoM)
		printf( "mopped up %d out of %d\n", mCount, num2MopUp );
}
