/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 6, 2004
LipidBilayer.cpp: implementation of the LipidBilayer class.

**************************************************************************/

#include "LipidBilayer.h"
#include "Monomer.h"
#include "Msg.h"
#include "BacteriumApp.h"
#include "BactEnvPdu.h"
#include <stdio.h>
#include <stdlib.h>

// Small molecule alphabet
// the default alphabet from which to construct small molecules
int LipidBilayer::alphabet[] = {
Lip1MN, Lip2MN, NullMN };
int LipidBilayer::lenAlphabet = 2;

// Motor Calibration
const int startMotorCalib = 50000; // time step to start recording runs&tumbles
const int endMotorCalib  = 200000; // time step to end recording runs&tumbles

// Only send PDU every nth time step.
// This is to allow some time for adaptation to occur.
const int txPduInterval = 1;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LipidBilayer::LipidBilayer()
{
	timeStepCounter = 1;
}

LipidBilayer::LipidBilayer( Grid *g, int xx, int yy, int r, int f, int gl, Polymer* c, bool a, int h, int w )
{
	grid = g;
	gridSizeX = grid->getNumCols();
	gridSizeY = grid->getNumRows();
	totalRuns = 0.0;
	totalTumbles = 0.0;
	currentCCW = 0;
	timeStepCounter = 1;
	outQ = NULL;
	windowID = 0;
	setX(xx);
	setY(yy);
	setRot(r);
	setFlip(f);
	setGridLayer(gl);
	setContainer(c);
	setAttached(a);
	setShape(h,w);
	setProcessRunsTumbles(false, NULL, 0);
#ifdef MOTOR_CALIBRATION
	// Motor calibration
	totalRunsCalib = 0.0;
	totalTumblesCalib = 0.0;
	timeStepCounterCalib = 1;
#endif
}

LipidBilayer::~LipidBilayer()
{

}

///////////////////////////////////////////
// setShape
// Set the 2D shape of this lipid bilayer.
// A lipid bilayer is a rectangle.
///////////////////////////////////////////
void LipidBilayer::setShape( int nRows, int nCols )
{
	int row, col;
	numRows = nRows;
	numCols = nCols;
	if (shape != NULL)
		delete shape;
	shape = new int[(numCols * numRows) + 1];
	memsetInt(shape, WaterMN, (numCols * numRows) + 1); // water fill
	// Two outermost positions in border of lipid bilayer contain lipids.
	for (row = 0; row < numRows; row++) {
		for (col = 0; col < numCols; col++) {
			if (row < 2 || row > numRows-3) {
				shape[row*numCols+col] = Lip1MN;
			}
			else if (col < 2 || col > numCols-3) {
				shape[row*numCols+col] = Lip1MN;
			}
		}
	}

	// Initialize bindings array
	if (bindings != NULL)
		delete bindings;
	bindings = new int[(numCols * numRows) + 1];
	memsetInt(bindings, NullRUL, (numCols * numRows) + 1);
	// Initialize rules array
	if (rules != NULL)
		delete rules;
	rules = (ruleEnum*) new int[(numCols * numRows) + 1];
	memsetInt((int*)rules, NullRUL, (numCols * numRows) + 1);
}

////////////////////////////////////////
// act
// Perform some action every time step.
//   Sum the current states of all six motors.
//   Send this result to various places:
//   - to the console window using a printf()
//   - to the EnvironmentApp if it's been created
//   - to a statistics file if stats is enabled
////////////////////////////////////////
void LipidBilayer::act( void )
{
	//printf( "LipidBilayer::act()\n" );
	BactEnvPdu *bep;

	// Perform this processing only if this is the membrane that surrounds a bacterium.
	if (processRunsTumbles) {
	bep = new BactEnvPdu();
	if (currentCCW == 6) { // number of motors is 6
		// eColi should move forward one unit if all motors are turning CCW
		bep->pduType = RunBPT;
		totalRuns++;
#ifdef MOTOR_CALIBRATION
		if (timeStepCounterCalib > startMotorCalib)
			totalRunsCalib++;
#endif
	}
	else { // at least one motor is turning CW
		// eColi should rotate randomly by 90/180/270 degrees
		bep->pduType = TumbleBPT;
		totalTumbles++;
#ifdef MOTOR_CALIBRATION
		if (timeStepCounterCalib > startMotorCalib)
			totalTumblesCalib++;
#endif
	}
	// Send PDU to optional EnvironmentApp; only every nth time step
	if ((timeStepCounter % txPduInterval) == 0) {
		if (outQ->enqueue( bep ) == Q_FULL)
			delete bep; // if Q is full then delete it
	}
	else
		delete bep; // not sent so delete it

	currentCCW = 0;

	// average proportion of runs should be about .606 = .92^6
	// under default conditions with no chemical messages from environment
	if (timeStepCounter == 1000) {
		printf( "WinID: %d Runs: %.1f Tumbles: %.1f ProportionRuns: %.5f\n",
		windowID, totalRuns, totalTumbles, totalRuns/(totalRuns+totalTumbles) );
		plotPropRuns( totalRuns, totalTumbles ); // write to stats file
		totalRuns = 0;
		totalTumbles = 0;
		timeStepCounter = 0;
	}
#ifdef MOTOR_CALIBRATION
		if (timeStepCounterCalib == endMotorCalib) {
			printf( "Motor Calibration: %d Runs: %.1f Tumbles: %.1f Proportion Runs: %.5f\n",
				windowID, totalRunsCalib,
				totalTumblesCalib,
				totalRunsCalib/(totalRunsCalib+totalTumblesCalib) );
		}
		timeStepCounterCalib++;
#endif
		timeStepCounter++;

	} // end if(processRunsTumbles)
}

//////////////////////////////////////////////
// plotPropRuns
// Write a line of data to a statistics file.
//////////////////////////////////////////////
void LipidBilayer::plotPropRuns( double numRuns, double numTumbles )
{
	// these files never get closed ?
	if (!(grid->stats->statsEnabled)) return;
	char fileName[100];
	static FILE* fp1 = NULL;
	static FILE* fp2 = NULL;
	if (windowID == 1) {
		if (fp1 == NULL) {
			// open file if not yet opened
			sprintf( fileName, "%s%c%s_%s.csv",
				grid->stats->getPathName(), statsSeries, grid->stats->getDateTime(), "runs1" );
			fp1 = fopen( fileName, "w+" );
			grid->stats->outputGnuplotScriptR( "runs1" );
		}
		fprintf( fp1, "%.5f\n", numRuns/(numRuns+numTumbles ) );
		fflush( fp1 );
	}
	else if (windowID == 2) {
		if (fp2 == NULL) {
			sprintf( fileName, "%s%c%s_%s.csv",
				grid->stats->getPathName(), statsSeries, grid->stats->getDateTime(), "runs2" );
			fp2 = fopen( fileName, "w+" );
			grid->stats->outputGnuplotScriptR( "runs2" );
		}
		fprintf( fp2, "%.5f\n", numRuns/(numRuns+numTumbles ) );
		fflush( fp2 );
	}
}

/////////////////////////////////////////
// move
// At present lipid bilayers don't move.
/////////////////////////////////////////
void LipidBilayer::move( void )
{
	//printf( "LipidBilayer::move()\n" );
}

///////////////////////////////////////
// E Coli; runs, tumbles, CCW movement
///////////////////////////////////////
void LipidBilayer::incCurrentCCW( void ) {currentCCW++;}
double LipidBilayer::getTotalRuns( void ) {return totalRuns;}
double LipidBilayer::getTotalTumbles( void ) {return totalTumbles;}
void LipidBilayer::setProcessRunsTumbles( bool prt, Queue *q, int winID )
{
	processRunsTumbles = prt;
	outQ = q;
	windowID = winID;
}

/////////////
// inventory
/////////////
void LipidBilayer::inventory( FILE *fp, Monomers *mn )
{
	if (getGhost())
		fprintf( fp, "Lg,%5d,%d,%d\n", id, numRows, numCols );
	else
		fprintf( fp, "L,%5d,%d,%d\n", id, numRows, numCols );
}


///////////
// doStats
///////////
void LipidBilayer::doStats( void )
{
}
