/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 5, 2004
Polymer.cpp: implementation of the Polymer class.

Polymer is an abstract class.
Many of the functions in this class need to be fully implemented in
concrete subclasses, and therefore have little or no code associated
with them in this file.
**************************************************************************/

#include "Polymer.h"
#include "Monomer.h"
#include "Grid.h"
#include "Msg.h"
#include "BacteriumApp.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

// A table used to find the polymer's type for statistics purposes.
key Polymer::keytab[] = {
	"TAR", Mcp_ST,
	"TSR", Mcp_ST,
	"TAP", Mcp_ST,
	"TRG", Mcp_ST,
	"CheW", CheW_ST,
	"CheA", CheA_ST,
	"CheY", CheY_ST,
	"CheB", CheB_ST,
	"CheR", CheR_ST,
	"CheZ", CheZ_ST,
	"Aspartate", Ligands_ST,
	"Serine", Ligands_ST,
	"Dipeptide", Ligands_ST,
	"Ribose_Galactose", Ligands_ST,
	"Axp", PhosFree_ST,
	"Meth", MethFree_ST,
	"Meth0", MethFree_ST,
	"MotorR1", Motor_ST,
	"MotorR2", Motor_ST,
	"MotorR3", Motor_ST,
	"MotorL1", Motor_ST,
	"MotorL2", Motor_ST,
	"MotorL3", Motor_ST
};
int Polymer::max_keys = 23; // number of values in the table

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Polymer::Polymer()
{
	gridSizeX = 0;
	gridSizeY = 0;
	hydrophobic = false;
	shape = NULL;
	bindings = NULL;
	rules = NULL;
	numRows = 0;
	numCols = 0;
	rotOriginX = 0;
	rotOriginY = 0;
	x = 0;
	y = 0;
	rotated = Rot0;
	flipped = FlipNone;
	id = 0;
	setName( "Polly" );
	gridLayer = NullLY;
	container = NULL;
	attached = false;
	changed = false;
	ghost = false;
	statsType = Null_ST;
}

Polymer::~Polymer()
{
	if (shape != NULL)
		delete shape;
	if (bindings != NULL)
		delete bindings;
	if (rules != NULL)
		delete rules;
}

/////////////////////////////////////////
// Global position set and get functions
/////////////////////////////////////////
void Polymer::setX( int xx ) {x = xx;} // x
int Polymer::getX( void )    {return x;}
void Polymer::setY( int yy ) {y = yy;} // y
int Polymer::getY( void )    {return y;}
void Polymer::setRot( int rot ) {rotated = rot;} // rotated
void Polymer::incRot( void ) {
	rotated += 90;
	if (rotated > Rot270)
		rotated = Rot0;
}
void Polymer::decRot( void ) {
	rotated -= 90;
	if (rotated < Rot0)
		rotated = Rot270;
}
int Polymer::getRot( void ) {return rotated;}
void Polymer::setFlip( int flip ) {flipped = flip;} // flipped
void Polymer::incFlip( void ) {
	if (++flipped > FlipBoth)
		flipped = FlipNone;
}
int Polymer::getFlip( void ) {return flipped;}

///////////////////////////////
// Other set and get functions
///////////////////////////////
void Polymer::setID( int iidd ) {id = iidd;}
int Polymer::getID( void ) {return id;}
void Polymer::setName( char* nm )
{
	assert (strlen(nm) <= MaxPolymerNameLen);
	strcpy( name, nm );
}
char* Polymer::getName( void ) {return name;}
void Polymer::setGridLayer( int gLayer ) {gridLayer = gLayer;}
int Polymer::getGridLayer( void ) {return gridLayer;}
void Polymer::setContainer( Polymer* cont ) {container = cont;}
Polymer* Polymer::getContainer( void ) {return container;}
void Polymer::setAttached( bool att ) {attached = att;}
bool Polymer::getAttached( void ) {return attached;}
void Polymer::setChanged( bool ch ) {changed = ch;}
bool Polymer::getChanged( void ) {return changed;}
void Polymer::setGhost( bool gh )
{
	ghost = gh;
}
bool Polymer::getGhost( void ) {return ghost;}
// report whether this polymer has a hydrophobic domain
bool Polymer::getHydrophobic() {return hydrophobic;}

// Set the stats type based on the name
void Polymer::setStatsType( char* polyName )
{
	for (int ix = 0; ix < max_keys; ix++) {
		if (strcmp( keytab[ix].word, polyName ) == NULL ) { // matched
			statsType = keytab[ix].scode;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// Shape set and get functions
// These functions are a good place to capture bugs involving array overruns,
// so there are lots of assert() statements.
//////////////////////////////////////////////////////////////////////////////
void Polymer::setShape( int* sh, int nRows, int nCols )
{
	assert( sh != NULL );
	shape = sh;
	numRows = nRows;
	numCols = nCols;
}
int* Polymer::getShape( void ) {return( shape );}
void Polymer::setShapeVal( int row, int col, int val )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	assert( val >= NullMN );
	assert( val < SizeOfMonomerEnumMN );
	shape[row*numCols + col] = val;
}

int Polymer::getShapeVal( int row, int col )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	assert( shape[row*numCols + col] >= NullMN );
	assert( shape[row*numCols + col] < SizeOfMonomerEnumMN );
	return shape[row*numCols + col];
}

//////////////////////////////////
// Bindings set and get functions
//////////////////////////////////
void Polymer::setBindings( int* b ) // assume nRows, nCols already set by setShape()
{
	assert( b != NULL );
	bindings = b;
}

int* Polymer::getBindings( void ) {return( bindings );}

void Polymer::setBindingVal( int row, int col, int val )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	assert( val >= NullBNG );
	bindings[row*numCols + col] = val;
}

int Polymer::getBindingVal( int row, int col )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	return bindings[row*numCols + col];
}

///////////////////////////////
// Rules get and set functions
///////////////////////////////
void Polymer::setRules( enum ruleEnum* r ) // assume nRows, nCols already set by setShape()
{
	rules = r;
}
enum ruleEnum* Polymer::getRules( void ) {return( rules );}

// place a specific rule in the rules array
// if that position already contains a rule, create a compund rule
void Polymer::setRuleVal( int row, int col, enum ruleEnum val )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	assert( val >= NullRUL );
	assert( val < SizeOfRuleEnum );
	int pos = row*numCols + col;
	if ((rules[pos] == NullRUL)
		|| (rules[pos] == SwapIfPhosphorylatedRUL) // these two just swap with each other
		|| (rules[pos] == SwapIfUnphosphorylatedRUL))
		rules[pos] = val;
	else
		rules[pos] = makeCompoundRule( rules[pos], val );
}
enum ruleEnum Polymer::getRuleVal( int row, int col )
{
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	return rules[row*numCols + col];
}

/////////////////////////////////////
// makeCompoundRule
// make a compound rule, if possible
// otherwise return existing rule
/////////////////////////////////////
enum ruleEnum Polymer::makeCompoundRule( enum ruleEnum existingRule, enum ruleEnum newRule )
{
	enum ruleEnum resultRule = existingRule;

	switch( existingRule ) {
	case ProteinMembraneBindingRUL:
		if (newRule == ProteinProteinBindingRUL)
			resultRule = PPandPMCompoundRUL;
		break;
	case ProteinProteinBindingRUL:
		if (newRule == AutoDephosphorylationRUL)
			resultRule = ADandPPCompoundRUL;
		break;
	default:
		break;
	}

	return resultRule;
}

///////////////////
// Get rows & cols
///////////////////
int Polymer::getNumRows( void ) {return( numRows );}
int Polymer::getNumCols( void ) {return( numCols );}

/////////////////////////////////////////
// act
// do a possible binding or other action
/////////////////////////////////////////
void Polymer::act( void )
{
	printf( "Polymer::act()\n" );
}

/////////////////////////////////
// move
// do a possible random movement
/////////////////////////////////
void Polymer::move( void )
{
	printf( "Polymer::move()\n" );
}

/////////////////////////////////////////////////////////////////
// doesOverlapMembrane
// Does this polymer overlap any membrane?
// search along the perimeter of the shape
// only check every second postion; membranes are 2 lipids thick
/////////////////////////////////////////////////////////////////
Polymer* Polymer::doesOverlapMembrane( void* gridV )
{
	Grid* grid = (Grid*)gridV;
	gint* mGrid = grid->getLayerMono( MembraneLY );
	gint* mPolyGrid = grid->getLayerPoly( MembraneLY );
	int row, col;
	int gridPos, contents;

	row = 0;
	for (col = 0; col < numCols; col += 2) {
		shape2grid( GposContS2G, row, col, &gridPos, &contents );
		if (mGrid[gridPos] == Lip1MN)
			return (grid->getPolymer(mPolyGrid[gridPos]));
	}
	if (numCols % 2 == 0) { // skipped last column if even number of columns
		shape2grid( GposContS2G, row, numCols-1, &gridPos, &contents );
		if (mGrid[gridPos] == Lip1MN)
			return (grid->getPolymer(mPolyGrid[gridPos]));
	}
	if (numRows < 2) return NULL;
	for (row = 2, col = 0; row < numRows; row += 2) {
		shape2grid( GposContS2G, row, col, &gridPos, &contents );
		if (mGrid[gridPos] == Lip1MN)
			return (grid->getPolymer(mPolyGrid[gridPos]));
	}
	for (col = 0, row = numRows-1; col < numCols; col += 2) {
		shape2grid( GposContS2G, row, col, &gridPos, &contents );
		if (mGrid[gridPos] == Lip1MN)
			return (grid->getPolymer(mPolyGrid[gridPos]));
	}
	for (row = 2, col = numCols-1; row < numRows; row += 2) {
		shape2grid( GposContS2G, row, col, &gridPos, &contents );
		if (mGrid[gridPos] == Lip1MN)
			return (grid->getPolymer(mPolyGrid[gridPos]));
	}
	return NULL;
}

///////////////
// doesOverlap
// Does this polymer overlap with another one.
///////////////
bool Polymer::doesOverlap( int myX, int myY, int myNCols, int myNRows, int opX, int opY, int opNCols, int opNRows, Polymer* op )
{
	if (op->getGridLayer() == MembraneLY) { // must be a LipidBilayer
		// only checking the case where protein/SM is supposed to stay inside membrane
		if ((myX == opX+1) || (myNCols+myX == opNCols+opX-1)
			|| (myY == opY+1) || (myNRows+myY == opNRows+opY-1))
			return true;
		return false;
	}
	else { // a protein or small molecule
		if ((myX >= opX)    && (myX <= opNCols+opX)
			&& (myY >= opY) && (myY <= opNRows+opY))
			return true;
		if ((myX >= opX)            && (myX <= opNCols+opX)
			&& (myNRows+myY >= opY) && (myNRows+myY <= opNRows+opY))
			return true;
		if ((myNCols+myX >= opX) && (myNCols+myX <= opNCols+opX)
			&& (myY >= opY)      && (myY <= opNRows+opY))
			return true;
		if ((myNCols+myX >= opX)    && (myNCols+myX <= opNCols+opX)
			&& (myNRows+myY >= opY) && (myNRows+myY <= opNRows+opY))
			return true;
		if ((opX > myX)    && (opX < myNCols+myX)
			&& (opY > myY) && (opY < myNRows+myY))
			return true;
		return false;
	}
}

////////////////////////////////////////////////////////////////
// printShape
// Print out the 2D monomer contents of a folded polymer shape.
////////////////////////////////////////////////////////////////
void Polymer::printShape( void )
{
	if (!Msg::infoM) return;
	char row, col;
	int mon; // monomer
	int printStyle = PrintCode;
	printf( "Folded Shape: %s\n", getName() );

	if (printStyle == PrintID) {
		for (row = numRows-1; row >= 0; row--) {
			for (col = 0; col < numCols; col++) {
				printf( "{%2d}", shape[row*numCols+col] );
			}
			printf( "\n" );
		}
	}
	else { // PrintCode
		printf( " +" );
		for (col = 0; col < numCols; col++)
			printf( "%c", '-' );
		printf( "+\n" );
		for (row = numRows-1; row >= 0; row--) {
			printf( "%c%c", (row%10 + '0'), '|' );
			for (col = 0; col < numCols; col++) {
				mon = getShapeVal( row, col );
				if (mon == NullMN)
					printf( "%c", '?' );
				else if (mon == WaterMN)
					printf( "%c", ' ' );
				else {
					printf( "%c", Monomers::get(mon)->code );
				}
			}
			printf( "%c%\n", '|' );
		}
		printf( " +" );
		for (col = 0; col < numCols; col++)
			printf( "%c", '-' );
		printf( "+\n" );

		// print column indices
		printf( "  " );
		for (col = 0; col < numCols; col++)
			printf( "%c", (col%10 + '0') );
		printf( " \n" );
	}
}

//////////////////////////////////////////
// memsetInt
// Fill an integer array with some value.
//////////////////////////////////////////
void Polymer::memsetInt( int* str, int fillInt, int len )
{
	int i;
	for (i = 0; i < len; i++)
		*(str+i) = fillInt;
	*(str+i) = NullMN;
}

//////////////////
// setHydrophobic
//////////////////
void Polymer::setHydrophobic()
{
}

/////////////////
// isHydrophobic
/////////////////
bool Polymer::isHydrophobic(int c)
{
	return false;
}

///////////////
// isAminoAcid
///////////////
bool Polymer::isAminoAcid( MonomerTypeEnum c )
{
	if (c>=AlaMN && c<=GluMeth4MN) //TyrPhosMN)
		return true;
	else
		return false;
}

///////////
// isLipid
///////////
bool Polymer::isLipid( MonomerTypeEnum c )
{
	if (c==Lip1MN || c==Lip2MN)
		return true;
	else
		return false;
}

//////////////////////////////////////////////////////////////
// getPistonOffset
// Get piston offset for a specified row number within shape.
//////////////////////////////////////////////////////////////
int Polymer::getPistonOffset(int rNum)
{
	return 0;
}

///////////////////
// setPistonOffset
///////////////////
void Polymer::setPistonOffset(int rNum, int val)
{

}

/////////////////////////////////////////////////////////////////
// grid2shape
// Convert global grid coordinates into local shape coordinates.
// do reverse of shape2grid()
// rType is not used for now
/////////////////////////////////////////////////////////////////
void Polymer::grid2shape( int rType, int gRow, int gCol, int* sRow, int* sCol, int* cont, int* bind )
{
	int col, row, colPrime, rowPrime;

	// Translation
	col = gCol - x;
	row = gRow - y;

	// Opposite transformations from shape2grid()
	switch( rotated ) {
	case Rot0:
		colPrime = col;
		rowPrime = row;
		break;
	case Rot90:
		// matrix 0 -1
		//        1  0
		colPrime = rotOriginX + (row-rotOriginY) * -1; // col * 0 + row * -1;
		rowPrime = rotOriginY + (col-rotOriginX);      // col * 1 + row *  0;
		break;
	case Rot180:
		// matrix -1  0
		//         0 -1
		colPrime = rotOriginX + (col-rotOriginX) * -1; // col * -1 + row *  0;
		rowPrime = rotOriginY + (row-rotOriginY) * -1; // col *  0 + row * -1;
		break;
	case Rot270:
		// matrix  0  1
		//        -1  0
		colPrime = rotOriginX + (row-rotOriginY);      // col *  0 + row * 1;
		rowPrime = rotOriginY + (col-rotOriginX) * -1; // col * -1 + row * 0;
		break;
	default:
		break;
	}
	colPrime -= getPistonOffset( rowPrime ); // only important for TMP proteins

	// same as shape2grid(); same operation twice produces the original
	switch( flipped) {
	case FlipNone:
		*sCol = colPrime;
		*sRow = rowPrime;
		break;
	case FlipVert:
		*sCol = colPrime;
		*sRow = numRows - rowPrime - 1;
		break;
	case FlipHorz:
		*sCol = numCols - colPrime - 1;
		*sRow = rowPrime;
		break;
	case FlipBoth:
		*sCol = numCols - colPrime - 1;
		*sRow = numRows - rowPrime - 1;
		break;
	default:
		break;
	}

	*cont = shape[*sRow * numCols + *sCol];
	*bind = bindings[*sRow * numCols + *sCol];
}

//////////////////////////////////////////////////////////////
// shape2grid
// Convert local shape coordinates into global grid position.
// also return contents of shape at those coordinates
// Inputs:
//   rType - type of output returned
//   sRow - shape row; range from 0 to numRows - 1
//   sCol - shape col; range from 0 to numCols - 1
// Outputs:
//   gridRow and GridCol
//   or grid position from 0 to gridSize - 1, and contents
//////////////////////////////////////////////////////////////
void Polymer::shape2grid( int rType, int sRow, int sCol, int* r1, int* r2 )
{
	int col, row, colPrime, rowPrime;

	// determine col within shape based on how it's flipped
	switch( flipped) {
	case FlipNone:
		col = sCol;
		row = sRow;
		break;
	case FlipVert:
		col = sCol;
		row = numRows - sRow - 1;
		break;
	case FlipHorz:
		col = numCols - sCol - 1;
		row = sRow;
		break;
	case FlipBoth:
		col = numCols - sCol - 1;
		row = numRows - sRow - 1;
		break;
	default:
		break;
	}
	col += getPistonOffset( sRow ); // only important for TMP proteins

	switch( rotated ) {
	case Rot0:
		colPrime = col;
		rowPrime = row;
		break;
	case Rot90:
		// matrix  0  1
		//        -1  0
		colPrime = rotOriginX + (row-rotOriginY);      // col *  0 + row * 1;
		rowPrime = rotOriginY + (col-rotOriginX) * -1; // col * -1 + row * 0;
		break;
	case Rot180:
		// matrix -1  0
		//         0 -1
		colPrime = rotOriginX + (col-rotOriginX) * -1; // col * -1 + row *  0;
		rowPrime = rotOriginY + (row-rotOriginY) * -1; // col *  0 + row * -1;
		break;
	case Rot270:
		// matrix 0 -1
		//        1  0
		colPrime = rotOriginX + (row-rotOriginY) * -1; // col * 0 + row * -1;
		rowPrime = rotOriginY + (col-rotOriginX);      // col * 1 + row *  0;
		break;
	default:
		break;
	}

	colPrime += x;
	rowPrime += y;
	if (rType == GyGxS2G) {
		*r1 = rowPrime; // grid row
		*r2 = colPrime; // grid col
	}
	else { // gposContS2G
		*r1 = rowPrime * gridSizeX + colPrime; // grid position
		*r2 = shape[sRow * numCols + sCol];    // contents
	}
}

/////////////////////
// setRotationOrigin
/////////////////////
void Polymer::setRotationOrigin( void )
{
	rotOriginX = (int)((float)(numCols-1) * 0.5f);
	rotOriginY = (int)((float)numRows * 0.5f);
}

////////////////////////////////////////
// doRule
// Do rules for all polymer subclasses.
////////////////////////////////////////
void Polymer::doRule( enum ruleEnum ruleNum, int row, int col )
{
	printf( "Polymer::doRule()\n" );
}

////////////////////////////////////////////////////////////////
// doRules
// Work through the entire polymer 2D shape checking for rules.
////////////////////////////////////////////////////////////////
void Polymer::doRules( void )
{
	int row, col;
	enum ruleEnum ruleNum;
	for (row = 0; row < numRows; row++) {
		for (col = 0; col < numCols; col++) {
			ruleNum = rules[row*numCols + col];
			if (ruleNum == NoMoreRulesRUL) return;
			if (ruleNum != NullRUL)
				doRule( ruleNum, row, col );
		}
	}
}

/////////////////////////////////////////////////////////////
// setNoMoreRules
// Place NoMoreRulesRUL in rules[] after last existing rule.
// to cut down on processing
/////////////////////////////////////////////////////////////
void Polymer::setNoMoreRules( void )
{
	int row, col;
	enum ruleEnum ruleNum;
	int noMoreRulesRow = 0;
	int noMoreRulesCol = -1;
	for (row = 0; row < numRows; row++) {
		for (col = 0; col < numCols; col++) {
			ruleNum = rules[row*numCols + col];
			if (ruleNum == NoMoreRulesRUL) return; // already set
			if (ruleNum != NullRUL) {
				noMoreRulesRow = row;
				noMoreRulesCol = col;
			}
		}
	}
	noMoreRulesCol++; // point to col after last rule
	if (noMoreRulesCol >= numCols) { // have gone beyond end of row
		noMoreRulesRow++;
		noMoreRulesCol = 0;
	}
	if (noMoreRulesRow < numRows) // have not gone beyond last row
		rules[noMoreRulesRow*numCols + noMoreRulesCol] = NoMoreRulesRUL;
}

///////////////////
// Phosphorylation
///////////////////
bool Polymer::isPhosphorylated( int row, int col ) {return false;}
void Polymer::phosphorylate( int row, int col ) {}
void Polymer::dephosphorylate( int row, int col ) {}

//////////////////////////////////////////////////////
// unbind
// Unbind all in bindings[] that match bindVal.
// set attached to false if there are now no bindings
//////////////////////////////////////////////////////
void Polymer::unbind( int bindVal )
{
	bool unattachFlag = true;
	int row, col;
	for (row = 0; row < numRows; row++) {
		for (col = 0; col < numCols; col++) {
			if (bindings[row*numCols + col] == bindVal)
				bindings[row*numCols + col] = NullBNG;
			else if (bindings[row*numCols + col] != NullBNG)
				unattachFlag = false;
		}
	}
	if (unattachFlag) // all positions are now NullBNG
		attached = false;
}

/////////////////////////////////////
// inventory
// Inventory - do nothing by default
/////////////////////////////////////
void Polymer::inventory( FILE *fp, Monomers *mn )
{
}

///////////
// doStats
///////////
void Polymer::doStats( void )
{
	printf( "s" );
}
