/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
July 12, 2004
Polymer.h: interface for the Polymer class.

Polymer is the abstract superclass for Protein, SmallMolecule, and LipidBilayer.
**************************************************************************/

#if !defined(AFX_POLYMER_H__7DC018C8_EB9B_4F67_AC34_011D6A7A390C__INCLUDED_)
#define AFX_POLYMER_H__7DC018C8_EB9B_4F67_AC34_011D6A7A390C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Monomer.h"
#include <stdio.h>

// Rotations - counterclockwise
const int Rot0     = 0; // default
const int Rot90   = 90;
const int Rot180 = 180;
const int Rot270 = 270;

// Flips
const int FlipNone = 0; // default
const int FlipVert = 1; // vertical
const int FlipHorz = 2; // horizontal
const int FlipBoth = 3; // vertical and horizontal

// Actions
typedef enum actionTypeEnum {
	MoveRightACT = 1,
	MoveLeftACT,
	MoveUpACT,
	MoveDownACT,
	RotateCwACT,
	RotateCcwACT,
	FlipVertACT,
	FlipHorzACT,
	
	SizeOfActionEnumACT // Size of enum
} actionTypeEnum;
const int DoNothingACT = 0;

// Bindings
const int NullBNG = -1; // or should it have some other value; -1 ?

// Rules
// for compound rules: try first rule, then second one
typedef enum ruleEnum {
	NullRUL = 0,
	ActivatePistonBindingRUL, // piston binding rules
	InhibitPistonBindingRUL,
	ProteinProteinBindingRUL, // protein-protein binding
	ProteinMembraneBindingRUL, // protein-membrane (lipid bilayer) binding
	PhosphorylationCheA3RUL, // phosphorylation rule specific to CheA domain 3
	SwapIfPhosphorylatedRUL, // pair of rules used for global conformation change (CheY)
	SwapIfUnphosphorylatedRUL,
	AutoPhosphorylationRUL, // used by CheA to autophosphorylate its own His monomer
	AutoDephosphorylationRUL, // used by CheY to autodephosphorylate its own monomer
	MotorRUL, // set the motor to turn CCW or CW
	MethylTransferaseRUL, // CheR
	MethylEsteraseRUL, // CheB
	PhosphataseRUL, // CheZ
	// compound rules
	PPandPMCompoundRUL, // ProteinProteinBindingRUL and then ProteinMembraneBindingRUL
	ADandPPCompoundRUL, // AutoDephosphorylationRUL and then ProteinProteinBindingRUL
	NoMoreRulesRUL, // put this after last rule in polymer; to save processing time
	SizeOfRuleEnum // Size of enum
} ruleEnum;

// shape2grid() return types
// used as inputs into the function to specify the type of data to be returned
const int GposContS2G = 1; // return gridPos and contents
const int GyGxS2G     = 2; // return gridY and gridX

const int MaxPolymerNameLen = 20;

// Print directives for printShape() and other print routines
const int PrintID = 1;   // print MonomerType::id (an integer)
const int PrintCode = 2; // print MonomerType::code (a character)

// used by setStatsType()
struct key {
	char *word;
	int scode;
};

// Polymer
class Polymer  
{
public:
	Polymer();
	virtual ~Polymer();

	// Set and get functions
	virtual void setPistonOffset( int rNum, int val );
	virtual int getPistonOffset( int rNum );
	bool getHydrophobic( void );
	// shape
	void setShape( int* sh, int nRows, int nCols );
	int* getShape( void );
	void setShapeVal( int row, int col, int val );
	int getShapeVal( int row, int col );
	// bindings
	void setBindings( int* b );
	int* getBindings( void );
	void setBindingVal( int row, int col, int val );
	int getBindingVal( int row, int col );
	void unbind( int bindVal );
	// rules
	void setRules( enum ruleEnum* r );
	enum ruleEnum* getRules( void );
	void setRuleVal( int row, int col, enum ruleEnum val );
	enum ruleEnum getRuleVal( int row, int col );
	enum ruleEnum makeCompoundRule( enum ruleEnum existingRule, enum ruleEnum newRule );
	// size of shape, bindings, rules
	int getNumRows( void );
	int getNumCols( void );

	// Global position set and get functions
	void setX( int xx ); // x
	int getX( void );
	void setY( int yy ); // y
	int getY( void );
	void setRot( int rot ); // rotated
	void incRot( void ); // increment
	void decRot( void ); // decrement
	int getRot( void );
	void setFlip( int flip ); // flipped
	void incFlip( void );
	int getFlip( void );

	// Other set and get functions
	void setID( int iidd );
	int getID( void );
	void setName( char* nm );
	char* getName( void );
	void setGridLayer( int gLayer ); // gLayer is one of values in GridLayerEnum
	int getGridLayer( void );
	void setContainer( Polymer* cont );
	Polymer* getContainer( void );
	void setAttached( bool att );
	bool getAttached( void );
	void setChanged( bool ch );
	bool getChanged( void );
	void setGhost( bool gh );
	bool getGhost( void );

	// Regular actions
	virtual void act( void );
	virtual void move( void );
	virtual void doRule( enum ruleEnum, int row, int col );
	void doRules( void );
	void setNoMoreRules( void );

	void printShape( void );

	// Convenience functions needed by various subclasses.
	// convert local polymer 2D shape coordinates to global grid coordinates
	void shape2grid( int rType, int sRow, int sCol, int* r1, int* r2 );
	// convert global grid coordinates to local polymer 2D shape coordinates
	void grid2shape( int rType, int gRow, int gCol, int* sRow, int* sCol, int* cont, int* bind );
	// calculate the center of the polymer; its rotation origin
	void setRotationOrigin( void );
	bool isPhosphorylated( int row, int col ); // is the monomer at this position phosphorylated?
	void phosphorylate( int row, int col ); // phosphorylate this monomer
	void dephosphorylate( int row, int col ); // dephosphorylate this monomer
	virtual bool isHydrophobic( int c); // is this monomer hydrophobic?
	bool isAminoAcid( enum MonomerTypeEnum c ); // is this monomer an amino acid?
	bool isLipid( enum MonomerTypeEnum c ); // is this monomer a lipid?

	// Inventory
	virtual void inventory( FILE *fp, Monomers *mn );
	// Stats
	virtual void doStats( void );

protected:
	int gridSizeX; // size of the global grid
	int gridSizeY;
	// Determine if this polymer (protein) has a hydrophobic domain
	virtual void setHydrophobic( void );

	bool hydrophobic; // whether or not this polymer has a hydrophobic domain
	int *shape; // 2D shape of the polymer; consists of monomers
	int *bindings; // binding for each monomer in the shape
	enum ruleEnum *rules; // possible rule for each monomer in the shape
	int numRows; // number of rows in the shape (folded protein string)
	int numCols; // number of columns in the shape (folded protein string)
	int rotOriginX; // rotation origin within the shape
	int rotOriginY;

	// Global position info (within the overall grid)
	int x;
	int y;
	int rotated;
	int flipped;

	int id; // unique ID of this polymer, as assigned by Grid
	char name[MaxPolymerNameLen]; // some identifying name; use is optional
	int gridLayer; // which layer in the grid this polymer belongs to
	Polymer* container; // which membrane or other container it belongs to

	// whether this polymer is attached to another polymer, that is ultimately
	// attached to something fixed such as a membrane
	bool attached;
	bool changed; // flag to indicate that something has changed relevant to the bindings
	bool ghost; // not currently part of the sim world; waiting to be activated

	// Statistics
	int statsType; // the type of this polymer for statistics purposes
	void setStatsType( char* polyName );
	static key keytab[];
	static int max_keys;

	// convenience function to fill an integer string with some default value
	void memsetInt( int* str, int fillInt, int len );

	// Movement actions
	void moveRight( void );
	void moveLeft( void );
	void moveUp( void );
	void moveDown( void );
	void rotateCw( void );
	void rotateCcw( void );
	void flipVert( void );
	void flipHorz( void );

	// Determine if this polymer overlaps something else
	// won't compile if #include "Grid.h", so have to make it void*
	Polymer* doesOverlapMembrane( void* gridV );
	bool doesOverlap( int myX, int myY, int myNCols, int myNRows,
		int opX, int opY, int opNCols, int opNRows, Polymer* op );
};

#endif // !defined(AFX_POLYMER_H__7DC018C8_EB9B_4F67_AC34_011D6A7A390C__INCLUDED_)
