/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 11, 2004
ProteinFactory.cpp: implementation of the ProteinFactory class.

**************************************************************************/

#include "ProteinFactory.h"
#include "BacteriumApp.h"

// Table of unfolded protein strings, used for initializing protein instances
char *ProteinFactory::protStr[SizeOfProteinEnumPT] = {
"DDDDhhhhEEDDGR**G**KGRHNGHNSD", // tar
"SSSShhhhEEDDGR**G**KGRHNGHNSD", // tsr
"WHRDhhhhEEDDGR**G**KGRHNGHNSD", // tap
"NSHWhhhhEEDDGR**G**KGRHNGHNSD", // trg
"HNSDGWR", // cheW
//"HRNRG\\HSSSDGDRKR", // cheA pre-phosphorylated
"HRNRGHSSSDGDRKR", // cheA
"HRNRGYLNK", // cheY
"HRNRGEDED", // cheB
"EEDD", // cheR
"YNLK", // cheZ
"KLNYhhhhDGShhhhYNLKGKLNYhhhhDSKR", // motor
"H\\HS\\ST\\TY\\Y" // test phosphorylated protein
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ProteinFactory::ProteinFactory( BacteriumApp *b )
{
	ba = b;
}

ProteinFactory::~ProteinFactory()
{

}

// Create an instance of the specified protein type
Protein* ProteinFactory::create( enum proteinTypeEnum pt, int x, int y, char *name )
{
	switch (pt) {

	// Receptors (source: Falke, 1997, Fig.6)
	// 4 (5 or 6) different periplasmic sensory domains "XXXX...."
	// transmembrane signaling domain "XXXXhhhh...."
	//   transmembrane region
	// cytoplasmic domain; highly conserved; each "binds the same CheW and CheA proteins" (p.482)
	//   linker region
	//   methylation region
	//   signaling domain
	case Tar_PT:
		// Transmembrane protein (TAR aspartate receptor)
		prot = new Protein(ba->grid, x, y, Rot90, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[Tar_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
		ba->grid->addPolymer( prot );
		break;
	case Tsr_PT:
		// Transmembrane protein (TSR serine receptor)
		prot = new Protein(ba->grid, x, y, Rot90, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[Tsr_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
		ba->grid->addPolymer( prot );
		break;
	case Tap_PT:
		// Transmembrane protein (TAP dipeptide receptor)
		prot = new Protein(ba->grid, x, y, Rot90, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[Tap_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
		ba->grid->addPolymer( prot );
		break;
	case Trg_PT:
		// Transmembrane protein (TRG ribose & galactose receptor)
		prot = new Protein(ba->grid, x, y, Rot90, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[Trg_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ActivatePistonBindingRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheW_PT:
		// Coupling protein CheW (source: Falke, 1997, p.489)
		// "possesses no known regulatory or catalytic function"
		// "a simple scaffold protein that couples CheA to the signaling module of the receptor"
		prot = new Protein(ba->grid, x, y, Rot90, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[CheW_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ProteinProteinBindingRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheA_PT:
		// Histidine Kinase CheA
		// "phosphorylation of His48"
		// Domains:
		// 1. P1 Phosphotransfer domain; contains His; passive; donates His to CheY
		// 2. P2 Response-regulator docking domain; docking site for CheY/CheB
		// 3. Catalytic domain; phosphorylates P1
		// 4. Receptor docking region; Tar binds here permanently
		prot = new Protein(ba->grid, x, y, Rot0, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[CheA_PT], ba->prob, name);
		prot->setRuleVal( 2, 1, ProteinProteinBindingRUL );
		prot->setRuleVal( 1, 1, PhosphorylationCheA3RUL );
		prot->setRuleVal( 1, 5, AutoPhosphorylationRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheY_PT:
		// CheY
		// binds to CheA at binding site HRNR
		prot = new Protein(ba->grid, x, y, Rot0, FlipVert, Protein2LY,
			ba->lipBilayer, false, protStr[CheY_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ProteinProteinBindingRUL );
		prot->setRuleVal( 1, 1, SwapIfPhosphorylatedRUL );
		prot->setRuleVal( 1, 0, ProteinProteinBindingRUL );
		prot->setRuleVal( 0, 0, AutoDephosphorylationRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheB_PT:
		// CheB
		// like CheY, it also binds to CheA at binding site HRNR
		prot = new Protein(ba->grid, x, y, Rot0, FlipVert, Protein2LY,
			ba->lipBilayer, false, protStr[CheB_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, ProteinProteinBindingRUL );
		prot->setRuleVal( 1, 1, SwapIfPhosphorylatedRUL );
		prot->setRuleVal( 1, 3, MethylEsteraseRUL );
		prot->setRuleVal( 0, 0, AutoDephosphorylationRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheR_PT:
		// CheR
		prot = new Protein(ba->grid, x, y, Rot0, FlipVert, Protein2LY,
			ba->lipBilayer, false, protStr[CheR_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, MethylTransferaseRUL );
		ba->grid->addPolymer( prot );
		break;

	case CheZ_PT:
		// CheZ
		prot = new Protein(ba->grid, x, y, Rot0, FlipVert, Protein2LY,
			ba->lipBilayer, false, protStr[CheZ_PT], ba->prob, name);
		prot->setRuleVal( 0, 0, PhosphataseRUL );
		ba->grid->addPolymer( prot );
		break;

	case Motor_PT:
		// Motor complex; all as one protein for now
		prot = new Protein(ba->grid, x, y, Rot0, FlipNone, Protein1LY,
			ba->lipBilayer, false, protStr[Motor_PT], ba->prob, name);
		prot->setRuleVal( 1, 1, MotorRUL );
		ba->grid->addPolymer( prot );
		break;

	case PhosTest_PT:
		// test phosphorylated protein
		prot = new Protein(ba->grid, x, y, Rot0, FlipNone, Protein1LY,
			ba->lipBilayer, NULL, protStr[PhosTest_PT], ba->prob, name);
		ba->grid->addPolymer( prot );
		break;

	default:
		break;
	}
	prot->printString();
	prot->printShape();
	return prot;
}
