/****************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 11, 2004
Statistics.cpp: implementation of the Statistics class.

The files output by the statistics functions are of several types:
1. Data files in comma separated format (CSV)
2. gnuplot script files that can read the data files and produce graphs
3. notes files for typing additional information about a specific trial

gnuplot is an open source application that plots graphs from data.
The PBNN system uses its ability to create these plots as PNG graphic files.
****************************************************************************/

#include "Statistics.h"
#include "Msg.h"
#include <stdlib.h>
#include <time.h>
#include <assert.h>

// Strings to print as column headers
char* Statistics::head[] = {
	"Null",
	"Mcp",
	"Piston",
	"Methed",
	"CheW",
	"CheA",
	"APhos",
	"AYb",
	"CheY",
	"YPhos",
	"YA",
	"YMot",
	"Motor",
	"MotY",
	"CheB",
	"BPhos",
	"BA",
	"CheR",
	"CheZ",
	"Lig",
	"PhosF",
	"MethF"
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Statistics::Statistics( char* pn, char* fnp, Probabilities *pr )
{
	fileNamePrefix = fnp;
	pathName = pn;
	prob = pr;
	fp = NULL;
	fpPlt = NULL;
	firstTime = true;
	statsEnabled = true;
	setDateTime();
	if (statsEnabled) {
		sprintf( dataFileName, "%s%c%s_%s.csv", pathName, statsSeries, dateTime, fileNamePrefix );
		char *mode = "w+";
		if ((fp = fopen(dataFileName, mode)) == NULL) {
			if (Msg::errorM) printf( "Can't open file: %s in %s mode\n", dataFileName, mode);
			exit(0);
		}
	}
	init();
	// establish mask; 1 indicates it will be printed out in stats file
	//   see StatsTypeEnum in .h file for order of stats columns corresponding to the mask items
	//      nMMMwAAAyyyyMMbbbRzLpM
	mask = "0011001001010001000000";
	outputHeadings();
}

Statistics::~Statistics()
{
	fclose( fp );
}

//////////////////////////////////////////////////
// init
// Inititialize stats counter array to all zeros.
//////////////////////////////////////////////////
void Statistics::init( void )
{
	for (int i = 0; i < SizeOfStatsTypeEnum_ST; i++)
		cnt[i] = 0;
}

/////////////////////////
// Set and Get functions
/////////////////////////
char* Statistics::getPathName( void )
{ return pathName; }
char* Statistics::getDateTime( void )
{ return dateTime; }
void Statistics::setMask( char* m )
{ mask = m; }
char* Statistics::getMask( void )
{ return mask; }

////////////////////////////////////////////
// output
// Output a statistics record to data file.
////////////////////////////////////////////
void Statistics::output( void )
{
	if (!statsEnabled) return;
	// First time
	if (firstTime) { // print header information at top of file
		fprintf( fp, "# number of MCP: %d\n", cnt[Mcp_ST] );
		fprintf( fp, "# number of CheW: %d\n", cnt[CheW_ST] );
		fprintf( fp, "# number of CheA: %d\n", cnt[CheA_ST] );
		fprintf( fp, "# number of CheY: %d\n", cnt[CheY_ST] );
		fprintf( fp, "# number of Motors: %d\n", cnt[Motor_ST] );
		fprintf( fp, "# number of CheB: %d\n", cnt[CheB_ST] );
		fprintf( fp, "# number of CheR: %d\n", cnt[CheR_ST] );
		fprintf( fp, "# number of CheZ: %d\n", cnt[CheZ_ST] );
		fprintf( fp, "# number of Ligands: %d\n", cnt[Ligands_ST] );
		fprintf( fp, "# number of Phos: %d\n", cnt[PhosFree_ST] );
		fprintf( fp, "# number of Meth: %d\n", cnt[MethFree_ST] );
		fprintf( fp, "# ProbLigandUnbindPRI: %d\n", prob->pr[ProbLigandUnbindPRI] );
		fprintf( fp, "# ProbCheAAutoPhosPRI: %d\n", prob->pr[ProbCheAAutoPhosPRI] );
		fprintf( fp, "# ProbCheYAutoDephosPRI: %d\n", prob->pr[ProbCheYAutoDephosPRI] );
		fprintf( fp, "# ProbCheBAutoDephosPRI: %d\n", prob->pr[ProbCheBAutoDephosPRI] );
		fprintf( fp, "# ProbMotorCwPRI: %d\n", prob->pr[ProbMotorCwPRI] );
		fprintf( fp, "# ProbPistonBindToLigandPRI: %d\n", prob->pr[ProbPistonBindToLigandPRI] );
		fprintf( fp, "# ProbCheBdemethInactivePRI: %d\n", prob->pr[ProbCheBdemethInactivePRI] );
		fprintf( fp, "# ProbCheRmethActivePRI: %d\n", prob->pr[ProbCheRmethActivePRI] );

		// create a corresponding notes file where extra info can be typed
		createNotesFile( "notes" );
		// create gnuplot script file that will convert data into a graph
		outputGnuplotScriptS();
		firstTime = false;
	}
	// Print masked data at every time step
	for (int i = 0; i < SizeOfStatsTypeEnum_ST; i++) {
		if (mask[i] == '1')
			fprintf( fp, "%d,", cnt[i] );
	}
	fprintf( fp, "\n" );
	fflush( fp ); // fflush so other processes can synchronously read the data
	init(); // reinitialize
}

////////////////////////////////////////////
// outputHeadings
// Output a headings line to the data file.
// This will be useful if the data file is read into Excel or another application that
// can read and process comma separated data (CSV format).
// gnuplot ignores this
////////////////////////////////////////////
void Statistics::outputHeadings( void )
{
	if (!statsEnabled) return;
	fprintf( fp, "#" ); // put gnuplot comment character at start of line
	for (int i = 1; i < SizeOfStatsTypeEnum_ST; i++) {
		if (mask[i] == '1')
			fprintf( fp, "%s,", head[i] );
	}
	fprintf( fp, "\n" );
	fflush( fp );
}

/////////////////////////////////////////
// outputGnuplotScriptS
// Output a gnuplot script file - Stats.
/////////////////////////////////////////
void Statistics::outputGnuplotScriptS( void )
{
	if (!statsEnabled) return;
	FILE *fpPlt;
	char fileName[100];
	char *modeW = "w"; // Open to Write

	// open file
	sprintf( fileName, "%s%c%s_%s.plt", pathName, statsSeries, dateTime, fileNamePrefix );
	if ((fpPlt = fopen(fileName, modeW)) == NULL) {
		printf( "Can't open file: %s in %s mode\n", fileName, modeW);
		return;
	}
	// get dataFileName without the path portion
	sprintf( dataFileName, "%c%s_%s.csv", statsSeries, dateTime, fileNamePrefix );

	// write out the lines of the gnuplot file
	fprintf( fpPlt, "# EASY MSc Dissertation - Protein-Based Neural Networks\n" );
	fprintf( fpPlt, "# Statistics from C++ Simulation\n" );
	fprintf( fpPlt, "# Ken Webb\n" );
	fprintf( fpPlt, "# %s\n", fileName );
	fprintf( fpPlt, "reset\n" );
	sprintf( fileName, "%c%s_%s.png", statsSeries, dateTime, fileNamePrefix );
	fprintf( fpPlt, "set output \"%s\"\n", fileName );
	fprintf( fpPlt, "set terminal png size 1024,768\n" );
	fprintf( fpPlt, "set datafile separator \",\"\n" );
	fprintf( fpPlt, "set xlabel \"Time\"\n" );
	fprintf( fpPlt, "set ylabel \"Quantities\"\n" );
	fprintf( fpPlt, "unset ytics\n" );
	fprintf( fpPlt, "set y2tics border mirror 1\n" );
	fprintf( fpPlt, "set mxtics 5\n" );
	fprintf( fpPlt, "set border 13\n" );
	fprintf( fpPlt, "set title \"Protein states changing over time (%s)\"\n", dateTime );
	fprintf( fpPlt, "set key right box 3\n" );
	fprintf( fpPlt, "set xrange [0:200000]\n" );
	fprintf( fpPlt, "#set yrange [0:25]\n" );
	fprintf( fpPlt, "set label \"MCP: %d\" at screen 0.0, screen 0.975\n", cnt[Mcp_ST] );
	fprintf( fpPlt, "set label \"CheW: %d\" at screen 0.0, screen 0.950\n", cnt[CheW_ST] );
	fprintf( fpPlt, "set label \"CheA: %d\" at screen 0.0, screen 0.925\n", cnt[CheA_ST] );
	fprintf( fpPlt, "set label \"CheY: %d\" at screen 0.0, screen 0.900\n", cnt[CheY_ST] );
	fprintf( fpPlt, "set label \"Motors: %d\" at screen 0.0, screen 0.875\n", cnt[Motor_ST] );
	fprintf( fpPlt, "set label \"CheB: %d\" at screen 0.0, screen 0.850\n", cnt[CheB_ST] );
	fprintf( fpPlt, "set label \"CheR: %d\" at screen 0.0, screen 0.825\n", cnt[CheR_ST] );
	fprintf( fpPlt, "set label \"CheZ: %d\" at screen 0.0, screen 0.800\n", cnt[CheZ_ST] );
	fprintf( fpPlt, "set label \"Lig: %d\" at screen 0.0, screen 0.775\n", cnt[Ligands_ST] );
	fprintf( fpPlt, "set label \"Phos: %d\" at screen 0.0, screen 0.750\n", cnt[PhosFree_ST] );
	fprintf( fpPlt, "set label \"Meth: %d\" at screen 0.0, screen 0.725\n", cnt[MethFree_ST] );

	fprintf( fpPlt, "set label \"pLigUnb: %d\" at screen 0.0, screen 0.700\n",
		prob->pr[ProbLigandUnbindPRI] );

	fprintf( fpPlt, "set label \"pAAutoP: %d\" at screen 0.0, screen 0.675\n",
		prob->pr[ProbCheAAutoPhosPRI] );

	fprintf( fpPlt, "set label \"pYAutoD: %d\" at screen 0.0, screen 0.650\n",
		prob->pr[ProbCheYAutoDephosPRI] );

	fprintf( fpPlt, "set label \"pBAutoD: %d\" at screen 0.0, screen 0.625\n",
		prob->pr[ProbCheBAutoDephosPRI] );

	fprintf( fpPlt, "set label \"pMotorCw: %d\" at screen 0.0, screen 0.600\n",
		prob->pr[ProbMotorCwPRI] );

	fprintf( fpPlt, "set label \"pPist2Lig: %d\" at screen 0.0, screen 0.575\n",
		prob->pr[ProbPistonBindToLigandPRI] );

	fprintf( fpPlt, "plot '%s' using 1 title \"Piston\" with lines 1, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' using 2 title \"Methed\" with lines 2, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' using 3 title \"APhos\" with lines 3, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' using 4 title \"YPhos\" with lines 4, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' using 5 title \"YMot\" with lines 5, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' using 6 title \"BPhos\" with lines 6\n", dataFileName );
	fprintf( fpPlt, "pause -1 \"Hit return to exit\"\n" );

	fclose( fpPlt );
}

////////////////////////////////////
// outputEnvAppHist
// Output EnvironmentApp histogram.
////////////////////////////////////
void Statistics::outputEnvAppHist( int bactIx, int* ligCount, int nLig )
{
	if (!statsEnabled) return;
	int i;
	FILE *fpHist = NULL;
	char fileName[100];
	char *fnpEahist = "eahist";
	char *modeW = "w"; // Open to Write

	// get max y value, and various summary stats
	int maxY = 0;
	int totalPoints = 0;
	int totalUpperHalf = 0;
	int totalLowerHalf = 0;
	int centerLine = nLig / 2;
	for (i = 0; i < nLig; i++) {
		if (ligCount[i] > maxY)
			maxY = ligCount[i];
		totalPoints += ligCount[i];
		if (i >= centerLine)
			totalUpperHalf += ligCount[i];
		else // (i < centerLine)
			totalLowerHalf += ligCount[i];
	}

	// get mean
	// in control group, mean should be at around the center line
	// in experimental group, should be greater than center line
	int sum = 0;
	for (i = 0; i < nLig; i++) {
		sum += ligCount[i] * i;
	}
	int mean = sum / totalPoints;

	// open file
	sprintf( fileName, "%s%c%s_%s%d.plt\0", pathName, statsSeries, dateTime, fnpEahist, bactIx+1 );
	if ((fpHist = fopen(fileName, modeW)) == NULL) {
		printf( "Can't open file: %s in %s mode\n", fileName, modeW);
		return;
	}

	// get dataFileName without the path portion
	sprintf( dataFileName, "%c%s_%s%d.plt", statsSeries, dateTime, fnpEahist, bactIx+1 );

	// write out the lines of the gnuplot file
	fprintf( fpHist, "# EASY MSc Dissertation - Protein-Based Neural Networks\n" );
	fprintf( fpHist, "# Statistics from C++ Simulation\n" );
	fprintf( fpHist, "# Ken Webb\n" );
	fprintf( fpHist, "# %s\n", fileName );
	fprintf( fpHist, "reset\n" );
	sprintf( fileName, "%c%s_%s%d.png", statsSeries, dateTime, fnpEahist, bactIx+1 );
	fprintf( fpHist, "set output \"%s\"\n", fileName );
	fprintf( fpHist, "set terminal png size 1024,768\n" );
	fprintf( fpHist, "set xlabel \"Ligand Concentration\"\n" );
	fprintf( fpHist, "set ylabel \"Number of Time Steps spent in Regions with that Ligand Concentration\"\n" );
	fprintf( fpHist, "set xtics 0, 10\n" );
	fprintf( fpHist, "set mxtics 5\n" );
	fprintf( fpHist, "set mytics 5\n" );
	fprintf( fpHist, "set grid front\n" );
	fprintf( fpHist, "set title \"Ligand Histogram for Bacterium %d (%s)\"\n", bactIx+1, dateTime );
	fprintf( fpHist, "set key right box 3\n" );
	fprintf( fpHist, "set xrange [0:%d]\n", nLig );
	fprintf( fpHist, "set yrange [0:%d]\n", maxY );
	fprintf( fpHist, "set label \"totalDataPoints: %d\" at screen 0.08, screen 0.93\n", totalPoints );
	fprintf( fpHist, "set label \"centerLine: %d|%d\" at screen 0.08, screen 0.91\n", centerLine-1, centerLine );
	fprintf( fpHist, "set label \"totalUpperHalf:  %d\" at screen 0.08, screen 0.89\n", totalUpperHalf );
	fprintf( fpHist, "set label \"totalLowerHalf:  %d\" at screen 0.08, screen 0.87\n", totalLowerHalf );
	fprintf( fpHist, "set label \"mean: %d\" at screen 0.08, screen 0.85\n", mean );
	double pcent;
	pcent = (100.0 * totalUpperHalf) / totalPoints;
	fprintf( fpHist, "set label \" %.1f\" at screen 0.24, screen 0.89\n", pcent );
	pcent = (100.0 * totalLowerHalf) / totalPoints;
	fprintf( fpHist, "set label \" %.1f\" at screen 0.24, screen 0.87\n", pcent );
	fprintf( fpHist, "set style fill solid 1.0\n" );
	fprintf( fpHist, "plot '-' title \"NLig\" with boxes\n" );

	for (i = 0; i < nLig; i++)
		fprintf( fpHist, "%d\n", ligCount[i] );
	fprintf( fpHist, "e\n" );
	fprintf( fpHist, "pause -1 \"Hit return to exit\"\n" );

	fclose( fpHist );
}

////////////////////////////////////////////////////////////////////////////////
// setDateTime
// Set the date and time as of the start of the simulation run.
// All files have this as part of the file name, to identify these files as all
// belonging to the same run.
////////////////////////////////////////////////////////////////////////////////
void Statistics::setDateTime( void )
{
	// Get current time
	time_t tnow;
	struct tm *tmnow;
	time( &tnow );
	tmnow = localtime( &tnow );

	// save in MMDDYY_HHMMSS format
	sprintf( dateTime, "%02d%02d%02d_%02d%02d%02d\0",
		tmnow->tm_mon + 1, tmnow->tm_mday, tmnow->tm_year - 100,
		tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec );
}

///////////////////////////////////////
// outputGnuplotScriptR
// Output a gnuplot script file - Runs
///////////////////////////////////////
void Statistics::outputGnuplotScriptR( char* fnp )
{
	if (!statsEnabled) return;
	//FILE *fpPlt = NULL;
	char fileName[100];
	char fileNameShort[50];
	char *modeW = "w"; // Open to Write

	// open file
	sprintf( fileName, "%s%c%s_%s.plt\0", pathName, statsSeries, dateTime, fnp );
	if ((fpPlt = fopen(fileName, modeW)) == NULL) {
		printf( "Can't open file: %s in %s mode\n", fileName, modeW);
		return;
	}
	// get dataFileName without the path portion
	sprintf( dataFileName, "%c%s_%s.csv", statsSeries, dateTime, fnp );

	// write out the lines of the gnuplot file
	fprintf( fpPlt, "# EASY MSc Dissertation - Protein-Based Neural Networks\n" );
	fprintf( fpPlt, "# Runs from C++ Simulation\n" );
	fprintf( fpPlt, "# Ken Webb\n" );
	fprintf( fpPlt, "# %s\n", fileName );
	fprintf( fpPlt, "reset\n" );
	sprintf( fileNameShort, "%c%s_%s.png\0", statsSeries, dateTime, fnp );
	fprintf( fpPlt, "set output \"%s\"\n", fileNameShort );
	fprintf( fpPlt, "set terminal png size 1024,768\n" );
	fprintf( fpPlt, "set datafile separator \",\"\n" );
	fprintf( fpPlt, "set xlabel \"Time\"\n" );
	fprintf( fpPlt, "set ylabel \"Proportion of Runs\"\n" );
	fprintf( fpPlt, "set mxtics 5\n" );
	fprintf( fpPlt, "set title \"Bacterial Chemotaxis - Runs vs. Tumbles (%s)\"\n", dateTime );
	fprintf( fpPlt, "set key left box 3\n" );
	fprintf( fpPlt, "set xrange [0:200]\n" );
	fprintf( fpPlt, "set yrange [0.4:1.0]\n" );
	fprintf( fpPlt, "plot '%s' title \"Win1\" lt 1, \\\n", dataFileName );
	fprintf( fpPlt, "     '%s' smooth bezier notitle lt 1\n", dataFileName );
	fprintf( fpPlt, "pause -1 \"Hit return to exit\"\n" );

	assert (fpPlt != NULL );
	fclose( fpPlt ); // may crash
}

////////////////////////////////////////////
// createNotesFile
// Create a notes file for this experiment.
////////////////////////////////////////////
void Statistics::createNotesFile( char* fnp )
{
	if (!statsEnabled) return;
	FILE *fpNt;
	char fileName[100];
	char *modeW = "w"; // Open to Write

	// open file
	sprintf( fileName, "%s%c%s_%s_%s.txt", pathName, statsSeries, dateTime, fnp, fileNamePrefix );
	if ((fpNt = fopen(fileName, modeW)) == NULL) {
		printf( "Can't open file: %s in %s mode\n", fileName, modeW);
		return;
	}

	// write out the lines of the gnuplot file
	fprintf( fpNt, "EASY MSc Dissertation - Protein-Based Neural Networks\n" );
	fprintf( fpNt, "Notes from C++ Simulation\n" );
	fprintf( fpNt, "Ken Webb\n" );
	fprintf( fpNt, "Experiment: %s\n\n", dateTime );

	fprintf( fpNt, "# number of MCP: %d\n", cnt[Mcp_ST] );
	fprintf( fpNt, "# number of CheW: %d\n", cnt[CheW_ST] );
	fprintf( fpNt, "# number of CheA: %d\n", cnt[CheA_ST] );
	fprintf( fpNt, "# number of CheY: %d\n", cnt[CheY_ST] );
	fprintf( fpNt, "# number of Motors: %d\n", cnt[Motor_ST] );
	fprintf( fpNt, "# number of CheB: %d\n", cnt[CheB_ST] );
	fprintf( fpNt, "# number of CheR: %d\n", cnt[CheR_ST] );
	fprintf( fpNt, "# number of CheZ: %d\n", cnt[CheZ_ST] );
	fprintf( fpNt, "# number of Ligands: %d\n", cnt[Ligands_ST] );
	fprintf( fpNt, "# number of Phos: %d\n", cnt[PhosFree_ST] );
	fprintf( fpNt, "# number of Meth: %d\n", cnt[MethFree_ST] );
	fprintf( fpNt, "# ProbLigandUnbindPRI: %d\n", prob->pr[ProbLigandUnbindPRI] );
	fprintf( fpNt, "# ProbCheAAutoPhosPRI: %d\n", prob->pr[ProbCheAAutoPhosPRI] );
	fprintf( fpNt, "# ProbCheYAutoDephosPRI: %d\n", prob->pr[ProbCheYAutoDephosPRI] );
	fprintf( fpNt, "# ProbCheBAutoDephosPRI: %d\n", prob->pr[ProbCheBAutoDephosPRI] );
	fprintf( fpNt, "# ProbMotorCwPRI: %d\n", prob->pr[ProbMotorCwPRI] );
	fprintf( fpNt, "# ProbPistonBindToLigandPRI: %d\n", prob->pr[ProbPistonBindToLigandPRI] );
	fprintf( fpNt, "# ProbCheBdemethInactivePRI: %d\n", prob->pr[ProbCheBdemethInactivePRI] );
	fprintf( fpNt, "# ProbCheRmethActivePRI: %d\n", prob->pr[ProbCheRmethActivePRI] );
	fprintf( fpNt, "\n" );

	fclose( fpNt );
}
