/**************************************************************************
EASY MSc Dissertation - Project
Ken Webb
August 10, 2004
Protein.cpp: implementation of the Protein class.

This is a concrete subclass of Polymer.
**************************************************************************/

#include "Protein.h"
#include "LipidBilayer.h"
#include "Monomer.h"
#include "Msg.h"
#include "Rndm.h"
#include "BacteriumApp.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

// uncomment this if running logic gates rather than bacgterium application
//#define LOGIC_GATES_APP 1

// 20 amino acids; the default alphabet from which to construct strings
// put Glycine at end so it can optionally be ignored as a random option
int Protein::alphabet[] = {
AlaMN, CysMN, AspMN, GluMN, PheMN, HisMN, IleMN, LysMN, LeuMN,
MetMN, AsnMN, ProMN, GlnMN, ArgMN, SerMN, ThrMN, ValMN, TrpMN, TyrMN,
GlyMN,
NullMN
};
int Protein::lenAlphabet = 20; // length of the alphabet, minus the NullMN

// Specialized alphabets.
//"AVFPMIL";  // hydrophobic amino acids
int Protein::alphHydrophobic[] = { AlaMN, ValMN, PheMN, ProMN, MetMN, IleMN, LeuMN, NullMN };
//"DEKR";     // charged amino acids
int Protein::alphCharged[]     =  { AspMN, GluMN, LysMN, ArgMN, NullMN };
//"TYHCNQW"; // polar amino acids
int Protein::alphPolar[]       = { SerMN, ThrMN, TyrMN, HisMN, CysMN, AsnMN, GlnMN, TrpMN, NullMN };
//"G";        // glycine amino acids
int Protein::alphGlycine[]     = { GlyMN, NullMN };
int Protein::lenAlphHydrophobic = 7;
int Protein::lenAlphCharged     = 4;
int Protein::lenAlphPolar       = 8;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Protein::Protein() {} // don't use

Protein::Protein( Grid *g, int xx, int yy, int r, int f, int gl,
				 Polymer* c, bool a, char* tStr, Probabilities* pr, char* nm )
{
	grid = g;
	gridSizeX = grid->getNumCols();
	gridSizeY = grid->getNumRows();
	proteinString = NULL;
	lenString = 0;
	numG = 0;
	pistonOffset = NULL;
	setX(xx);
	setY(yy);
	setRot(r);
	setFlip(f);
	setGridLayer(gl);
	setContainer(c);
	setAttached(a);
	setString(tStr);
	setName( nm );
	setStatsType( nm );
	prob = pr;
}

Protein::~Protein()
{
	if (proteinString != NULL)
		delete proteinString;
}

////////////////////////////////////////////////////////
// createRandomString
// Create a random protein string of a specific length.
////////////////////////////////////////////////////////
void Protein::createRandomString( int len )
{
	int* templateString = new int[len+1]; // create temporary string
	memsetInt( templateString, WildCardMN, len ); // fill with wild cards
	setString( templateString, len ); // convert wild cards to specific random values
	delete templateString; // delete temporary string
}

////////////////
// setString
// Set protein string to its final unfolded value.
//////////////
void Protein::setString( int* templateStr, int len )
{
	int rNum;
	int i;

	if (proteinString != NULL)
		delete proteinString; // delete any existing string
	proteinString = new int[len+1]; // create a new blank string
	lenString = len;

	// replace all wild card characters with randomly chosen values
	for (i = 0; i < len; i++) {
		switch (templateStr[i]) {
		case WildCardMN:
			rNum = Rndm::getRandomInt( lenAlphabet-1 );
			if (rNum == WaterMN) rNum++; // quick fix
			proteinString[i] = alphabet[rNum];
			break;
		case WildCardNoGlyMN:
			rNum = Rndm::getRandomInt( lenAlphabet-2 ); // ignore Glycine
			if (rNum == WaterMN) rNum++; // quick fix
			proteinString[i] = alphabet[rNum];
			break;
		case HydrophobicMN:
			rNum = Rndm::getRandomInt( lenAlphHydrophobic-1 );
			proteinString[i] = alphHydrophobic[rNum];
			break;
		case ChargedMN:
			rNum = Rndm::getRandomInt( lenAlphCharged-1 );
			proteinString[i] = alphCharged[rNum];
			break;
		case PolarMN:
			rNum = Rndm::getRandomInt( lenAlphPolar-1 );
			proteinString[i] = alphPolar[rNum];
			break;
		// if not a wild card, keep what has been specified
		default:
			proteinString[i] = templateStr[i];
			break;
		}
		if (proteinString[i] == TurnMN)
			numG++;
	}
	proteinString[i] = NullMN;
	foldString(); // fold the string
	setHydrophobic(); // check if it has a hydrophobic binding domain
}

////////////////////////////////////////////////////////////////////////
// setString
// Convert char array into int array
// a convenience routine to allow user to specify as a character string
////////////////////////////////////////////////////////////////////////
void Protein::setString( char* charStr )
{
	int from = 0;
	int to = 0;
	int len = strlen( charStr );
	int* intStr = new int[len+1];
	while (from < len) {
		switch (charStr[from]) {
		case 'A': intStr[to] = AlaMN; break;
		case 'C': intStr[to] = CysMN; break;
		case 'D': intStr[to] = AspMN; break;
		case 'E': intStr[to] = GluMN; break;
		case 'F': intStr[to] = PheMN; break;
		case 'G': intStr[to] = GlyMN; break;
		case 'H': intStr[to] = HisMN; break;
		case 'I': intStr[to] = IleMN; break;
		case 'K': intStr[to] = LysMN; break;
		case 'L': intStr[to] = LeuMN; break;
		case 'M': intStr[to] = MetMN; break;
		case 'N': intStr[to] = AsnMN; break;
		case 'P': intStr[to] = ProMN; break;
		case 'Q':
		case 'q': intStr[to] = GlnMN; break;
		case 'R': intStr[to] = ArgMN; break;
		case 'S': intStr[to] = SerMN; break;
		case 'T': intStr[to] = ThrMN; break;
		case 'V': intStr[to] = ValMN; break;
		case 'W':
		case 'w': intStr[to] = TrpMN; break;
		case 'Y': intStr[to] = TyrMN; break;
		case '#': intStr[to] = WildCardMN; break;
		case '*': intStr[to] = WildCardNoGlyMN; break;
		case 'h': intStr[to] = HydrophobicMN; break;
		case 'c': intStr[to] = ChargedMN; break;
		case 'p': intStr[to] = PolarMN; break;
		case '\\': // phosphorylated
			from++;
			if (charStr[from] == 'H')
				intStr[to] = HisPhosMN;
			else if (charStr[from] == 'S')
				intStr[to] = SerPhosMN;
			else if (charStr[from] == 'T')
				intStr[to] = ThrPhosMN;
			else if (charStr[from] == 'Y')
				intStr[to] = TyrPhosMN;
			else {
				intStr[to] = NullMN;
				if (Msg::errorM) printf("ERROR: invalid phosphorylated\n");
			}
			break;
		default: intStr[to] = NullMN; break;
		}
		to++; from++;
	}
	intStr[to] = NullMN;
	setString( intStr, len ); // set to its final unfolded form
	delete intStr;
}

/////////////
// getString
/////////////
int* Protein::getString( void )
{
	return( proteinString );
}

///////////////////////////////
// foldString
// Fold string into a 2D form.
///////////////////////////////
void Protein::foldString( void )
{
	static int forwardDir = 1;
	static int backwardDir = 2;
	int direction = forwardDir;
	int ixFrom = 0;
	int numToRows = numG + 1;
	int numToCols = lenString * 2;
	int ixToRow = 0; // start in first row
	int ixToCol = numToCols / 2; // start in center of first row
	int i, jFrom, jTo;
	int aminoAcid;

	// initialize clipping region
	int minCol = ixToCol;
	int maxCol = ixToCol;

	int tempFoldedLen = numToRows * numToCols;
	int *tempFolded = new int[tempFoldedLen + 1];
	memsetInt(tempFolded, WaterMN, tempFoldedLen); // water fill

	// copy from 1D string into 2D folded structure
	// does first row in forward direction, next in reverse direction, and so on
	// TurnMN inicates it's time to turn creating a fold
	for (ixFrom = 0; ixFrom < lenString; ixFrom++) {
		if (direction == forwardDir) {
			tempFolded[ixToRow*numToCols + ixToCol] = proteinString[ixFrom];
			if (proteinString[ixFrom] == TurnMN) {
				direction = backwardDir;
				ixToRow++;
				if (ixToCol > maxCol)
					maxCol = ixToCol;
				ixToCol--; // leave TurnMM on its own at end of loop
			}
			else
				ixToCol++;
		}
		else { // direction == backwardDir
			tempFolded[ixToRow*numToCols + ixToCol] = proteinString[ixFrom];
			if (proteinString[ixFrom] == TurnMN) {
				direction = forwardDir;
				ixToRow++;
				if (ixToCol < minCol)
					minCol = ixToCol;
				ixToCol++; // leave TurnSYM on its own at end of loop
			}
			else
				ixToCol--;
		}
	}
	// Final line calculations if final character was not TurnSymbol
	if (ixToCol > maxCol)
		maxCol = ixToCol-1;
	if (ixToCol < minCol)
		minCol = ixToCol+1;

	// Clip to region bounded by minCol and maxCol
	numRows = numToRows;
	numCols = maxCol-minCol+1;
	// Compute the rotation origin which should be as close to center as possible
	setRotationOrigin();
	if (shape != NULL)
		delete shape;
	shape = new int[(numCols * numRows) + 1];
	memsetInt(shape, WaterMN, (numCols * numRows) + 1); // water fill
	// copy into final shape structure
	for (i = 0; i < numRows; i++) {
		for (jFrom = minCol, jTo = 0; jFrom < maxCol+1; jFrom++, jTo++) {
			aminoAcid = tempFolded[i*numToCols + jFrom];
			if (aminoAcid == NullMN)
				aminoAcid = WaterMN;
			shape[i*numCols + jTo] = aminoAcid;
		}
	}
	// crashes if delete when have OR string of len=4
	// probably OK now; various bugs that caused crashes have been fixed
	//delete tempFolded;

	// Set piston offsets to default value of 0
	pistonOffset = new int[numRows];
	for (i = 0; i < numRows; i++)
		pistonOffset[i] = 0; // default: 0

	// Initialize bindings array
	if (bindings != NULL)
		delete bindings;
	bindings = new int[(numCols * numRows) + 1];
	memsetInt(bindings, NullBNG, (numCols * numRows) + 1);

	// Initialize rules array
	if (rules != NULL)
		delete rules;
	rules = new enum ruleEnum[(numCols * numRows) + 1];
	memsetInt((int*)rules, NullRUL, (numCols * numRows) + 1);
}

//////////////////////////////////////////////////////
// printString
// Print the 1D protein string to the console window.
//////////////////////////////////////////////////////
void Protein::printString( void )
{
	if (!Msg::infoM) return;
	int i = 0;
	// print either the numeric ID, or the more concise single character code
	int printStyle = PrintCode;

	printf( "Protein String: %s\n", getName() );
	if (printStyle == PrintID) {
		while (proteinString[i] != NullMN) {
			printf( "[%2d]", proteinString[i]);
			i++;
		}
		printf( "\n" );
	}
	else { // PrintCode
		while (proteinString[i] != NullMN) {
			printf( "%c", Monomers::get(proteinString[i])->code );
			i++;
		}
		printf( "\n" );
	}
}

/////////////////////////////////////////
// act
// Do any rules defined for this protein.
/////////////////////////////////////////
void Protein::act( void )
{
	doRules();
}

//////////////////////////////////
// move
// Do a possible random movement.
//////////////////////////////////
void Protein::move( void )
{
	if (attached) return;
	int rNum = Rndm::getRandomInt( 1, SizeOfActionEnumACT-1 );
	switch (rNum) {
	case MoveRightACT:
		moveRight();
		break;
	case MoveLeftACT:
		moveLeft();
		break;
	case MoveUpACT:
		moveUp();
		break;
	case MoveDownACT:
		moveDown();
		break;
	case RotateCwACT:
		rotateCw();
		break;
	case RotateCcwACT:
		rotateCcw();
		break;
	case FlipVertACT:
		flipVert();
		break;
	case FlipHorzACT:
		flipHorz();
		break;
	default: // 0
		// safe to ignore this; protein does nothing during this time step
		// printf( "ERROR: Protein::act() default %d\n", rNum );
		break;
	}
}

////////////////////////////////////////////////////////////
// matchP_P
// Match monomers from two different proteins
// eventually use binding affinity table with probabilities
////////////////////////////////////////////////////////////
bool Protein::matchP_P( MonomerTypeEnum monP1, MonomerTypeEnum monP2 )
{
	if ((monP1 == MatchWaterMN) && (monP2 == WaterMN))
		return true;
	if (monP1 == monP2)
		return true;
	else
		return false;
}

/////////////////////////////////////////////////////////////
// matchP_SM
// Match protein monomer with a small molecule (SM) monomer.
/////////////////////////////////////////////////////////////
bool Protein::matchP_SM( MonomerTypeEnum monP, MonomerTypeEnum monSM )
{
	if (monP + (SmA_MN-AlaMN) == monSM)
		return true;
	else
		return false;
}

//////////////////////////////////////////////////////////////////////
//										Movement Actions            //
//////////////////////////////////////////////////////////////////////
// For all movements, it makes a tentative move, checks to see if the
// new position overlaps the membrane, and reverses the movement it
// it does overlap.
//////////////////////////////////////////////////////////////////////

/////////////
// moveRight
/////////////
void Protein::moveRight( void )
{
	Polymer* bilayer = NULL;
	if (x+1 < gridSizeX) {
		x++;
		bilayer = doesOverlapMembrane(grid);
		if (bilayer != NULL) {
			x--;
			// if is hydrophobic, then attach it to the membrane
			if ((getHydrophobic()) && (bilayer == container)) {
				// move it so it straddles the membrane
				// at right angles to its hydrophobic region
				flipped = FlipNone;
				rotated = Rot180;
				x = getHydroDiff( MoveRightACT, bilayer );
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
// getHydroDiff
// Calculate new position of protein such that its hydrophobic domain
// straddles the membrane.
//////////////////////////////////////////////////////////////////////
int Protein::getHydroDiff( int moveNum, Polymer* pol )
{
	int gyProtein, gyBilayer, gxProtein, gxBilayer;

	// convert local shape coordinates to global grid coordinates
	shape2grid( GyGxS2G, 0, hydrophobicCol, &gyProtein, &gxProtein );

	switch (moveNum) {
	case MoveRightACT:
		pol->shape2grid( GyGxS2G, 0, pol->getNumCols(), &gyBilayer, &gxBilayer );
		return (x + (gxBilayer - gxProtein));
	case MoveLeftACT:
		pol->shape2grid( GyGxS2G, 0, 0, &gyBilayer, &gxBilayer );
		return (x - (gxProtein - gxBilayer) - 1);
	case MoveUpACT:
		pol->shape2grid( GyGxS2G, pol->getNumRows(), 0, &gyBilayer, &gxBilayer );
		return (y + (gyBilayer - gyProtein));
	case MoveDownACT:
		pol->shape2grid( GyGxS2G, 0, 0, &gyBilayer, &gxBilayer );
		return (y - (gyProtein - gyBilayer) - 1);
	default:
		if (Msg::errorM) printf( "ERROR: getHydroDiff() default=%d\n", moveNum );
		return 0;
		break;
	}
}

////////////
// moveLeft
////////////
void Protein::moveLeft( void )
{
	Polymer* bilayer = NULL;
	if (x-1 >= 0) {
		x--;
		bilayer = doesOverlapMembrane(grid);
		if (bilayer != NULL) {
			x++;
			// if is hydrophobic, then attach it to the membrane
			if ((getHydrophobic()) && (bilayer == container)) {
				flipped = FlipNone;
				rotated = Rot0;
				x = getHydroDiff( MoveLeftACT, bilayer );
			}
		}
	}
}

//////////
// moveUp
//////////
void Protein::moveUp( void )
{
	Polymer* bilayer = NULL;
	if (y+1 < gridSizeY) {
		y++;
		bilayer = doesOverlapMembrane(grid);
		if (bilayer != NULL) {
			y--;
			// if is hydrophobic, then attach it to the membrane
			if ((getHydrophobic()) && (bilayer == container)) {
				flipped = FlipNone;
				rotated = Rot90;
				y = getHydroDiff( MoveUpACT, bilayer );
			}
		}
	}
}

////////////
// moveDown
////////////
void Protein::moveDown( void )
{
	Polymer* bilayer = NULL;
	if (y-1 >= 0) {
		y--;
		bilayer = doesOverlapMembrane(grid);
		if (bilayer != NULL) {
			y++;
			// if is hydrophobic, then attach it to the membrane
			if ((getHydrophobic()) && (bilayer == container)) {
				flipped = FlipNone;
				rotated = Rot270;
				y = getHydroDiff( MoveDownACT, bilayer );
			}
		}
	}
}

/////////////////////////////////
// rotateCw
// Rotate protein clockwise (CW)
/////////////////////////////////
void Protein::rotateCw( void )
{
	incRot();
	if (doesOverlapMembrane(grid)) {
		decRot();
	}
}

//////////////////////////////////////////
// rotateCcw
// Rotate protein counter clockwise (CCW)
//////////////////////////////////////////
void Protein::rotateCcw( void )
{
	decRot();
	if (doesOverlapMembrane(grid)) {
		incRot();
	}
}

////////////////////////////////
// flipVert
// Flip the protein vertically.
////////////////////////////////
void Protein::flipVert( void )
{
	incFlip();
}

//////////////////////////////////
// flipHorz
// Flip the protein horizontally.
//////////////////////////////////
void Protein::flipHorz( void )
{
	incFlip();
}

///////////////////////////////////////////////////////////////////////////
// setHydrophobic
// Determine if this protein has a hydrophobic domain. Defined as:
// - any rectangular area of size at least 1 x 4,
// - in which all amino acids are of type hydrophobic,
// - those contained within alphHydrophobic[],
// - any parallel substring must also be hydrophobic in same area
// The intent is that the hydrophobic domain can fit within a LipidBilayer
// membrane, in the direction of its covalent bonds.
// Since the actual structure of the protein never changes, this function
// only needs to be called once as soon as the protein is created.
///////////////////////////////////////////////////////////////////////////
void Protein::setHydrophobic()
{
	int col;
	int count = 0;
	for (col = 0; col < numCols; col++) {
		if (isColHydrophobic( col )) {
			count++;
			if (count == 4) {
				hydrophobic = true;
				hydrophobicCol = col - 3;
				rules[hydrophobicCol] = ProteinMembraneBindingRUL;
				break;
			}
		}
		else {
			count = 0; // reset
		}
	}
}

///////////////////////////////////////////
// isColHydrophobic
// Is entire column hydrophobic?
///////////////////////////////////////////
bool Protein::isColHydrophobic(int col)
{
	int row;
	for (row = 0; row < numRows; row++) {
		if (!isHydrophobic( shape[row*numCols+col] )
			&& (shape[row*numCols+col] != WaterMN))
			return false;
	}
	return true;
}

////////////////////////////////
// isHydrophobic
// Is this monomer hydrophobic?
////////////////////////////////
bool Protein::isHydrophobic(int c)
{
	int i;
	for (i = 0; i < lenAlphHydrophobic; i++) {
		if (alphHydrophobic[i] == c)
			return true;
	}
	return false;
}

//////////////////////////////////////////////////////////////
// getPistonOffset
// Get piston offset for a specified row number within shape.
// will be either 1 or 0; piston row of protein is extended or not
//////////////////////////////////////////////////////////////
int Protein::getPistonOffset(int rNum)
{
	return pistonOffset[rNum];
}

///////////////////
// setPistonOffset
///////////////////
void Protein::setPistonOffset(int rNum, int val)
{
	pistonOffset[rNum] = val;
}

/////////////////////////////////////////
// Probabilities - set and get functions
/////////////////////////////////////////
void Protein::setProbabilities( Probabilities* pr ) {prob = pr;}
Probabilities* Protein::getProbabilities( void ) {return prob;}
void Protein::setProbVal( int id, int val )
{
	if (id < SizeOfProbIdEnumPRI)
		prob->pr[id] = val;
}
int Protein::getProbVal( int id )
{
	if (id < SizeOfProbIdEnumPRI)
		return prob->pr[id];
	else
		return 0;
}

////////////////////////
// doRule
// Do one Protein rule.
// Pass in the rule number found at the row and column in the protein's 2D
// folded structure.
//   rRow: row where rule was found
//   rCol: col where rule was found
////////////////////////
void Protein::doRule( enum ruleEnum ruleNum, int row, int col )
{
	// do lots of assertion checking (debugging)
	assert( row >= 0 );
	assert( row < numRows );
	assert( col >= 0 );
	assert( col < numCols );
	assert( ruleNum > NullRUL );
	assert( ruleNum < SizeOfRuleEnum );

	switch (ruleNum) {

	case ActivatePistonBindingRUL:
	case InhibitPistonBindingRUL:
		doRule_PistonBinding( ruleNum, row, col );
		break;

	case ProteinProteinBindingRUL:
		doRule_ProteinProteinBinding( ruleNum, row, col );
		break;

	case ProteinMembraneBindingRUL:
		doRule_ProteinMembraneBinding( ruleNum, row, col );
		break;

	case PhosphorylationCheA3RUL:
		doRule_PhosphorylationCheA3( ruleNum, row, col );
		break;

	case SwapIfPhosphorylatedRUL:
	case SwapIfUnphosphorylatedRUL:
		doRule_SwapOnPhosChange( ruleNum, row, col );
		break;

	case AutoPhosphorylationRUL:
		doRule_AutoPhosphorylation( ruleNum, row, col );
		break;

	case AutoDephosphorylationRUL:
		doRule_AutoDephosphorylation( ruleNum, row, col );
		break;

	case MotorRUL:
		doRule_Motor( ruleNum, row, col );
		break;

	case MethylTransferaseRUL: // CheR
		doRule_MethylTransferase( ruleNum, row, col );
		break;

	case MethylEsteraseRUL: // CheB
		doRule_MethylEsterase( ruleNum, row, col );
		break;

	case PhosphataseRUL: // CheZ
		doRule_Phosphatase( ruleNum, row, col );
		break;

	// compound rule
	case PPandPMCompoundRUL:
		if (Msg::debugM) printf( "PPandPMCompoundRUL\n" );
		doRule_ProteinProteinBinding( ruleNum, row, col );
		doRule_ProteinMembraneBinding( ruleNum, row, col );
		break;

	// compound rule
	case ADandPPCompoundRUL:
		if (Msg::debugM) printf( "ADandPPCompoundRUL\n" );
		doRule_AutoDephosphorylation( ruleNum, row, col );
		doRule_ProteinProteinBinding( ruleNum, row, col );
		break;

	default:
		if (Msg::errorM) printf( "ERROR: Protein::doRule default=%d\n", ruleNum );
		break;
	}
}

/////////////////////////
// doRule_PistonBinding
// Piston Binding rule
/////////////////////////
void Protein::doRule_PistonBinding( enum ruleEnum ruleNum, int rRow, int rCol )
{
	int col = 0;
	int count = 0;
	int gridPos = 0;
	int contents = 0;
	int len = 4;
	int rNum = -1; // random number; initially < methLevel
	int methLevel = 0;
	int methCalcType = 2; // 1=linear, 2=exponential, 3=sigmoidal

	if (len > numCols) return;
	// check if the monomers are still bound, or have become unbound
	// should really be checking if corresponding monomers are still bound in other polymer?
	if (pistonOffset[rRow] == 1) {
		for (col = rCol; col < rCol+len; col++) {
			if (bindings[rRow*numCols + col] == NullBNG) {
				if (ruleNum == ActivatePistonBindingRUL) {
					pistonOffset[rRow] = 0;
					changed = true; // force execution of adjustBindings()
				}
				return;
			}
		}
		return;
	}

	// pistonOffset not yet activated
	gint* mGrid = grid->getLayerMono( SmallMolLY );
	gint* mPolyGrid = grid->getLayerPoly( SmallMolLY );
	int ligID = 0; // other protein
	Polymer *lig = NULL;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer

	for (col = rCol; col < rCol+len; col++) {
		shape2grid( GposContS2G, rRow, col, &gridPos, &contents );
		if (matchP_SM( (MonomerTypeEnum)contents,
			(MonomerTypeEnum)mGrid[gridPos] ))
			count++;
		else {
			count = 0;
		}
		if (count == 1)
			ligID = mPolyGrid[gridPos]; // this is the ligand I'm trying to match
		else if (count > 1) {
			if (mPolyGrid[gridPos] != ligID)
				count = 0;
		}
		if (count == len) break;
	}
	if (count == len) { // matched a ligand
		// effect of methylation on ligand binding
		// "high methylation level decreases the affinity of the receptor supramolecular
		//  complex to attractants" (Bren et al, 2000, p.6870)
		if (methCalcType == 1) { // linear
			methLevel = 8 - getSummedMethLevel(); // range: 0 to 8 inclusive
			rNum = Rndm::getRandomInt(prob->pr[ProbPistonBindToLigandPRI]); // range: 0 to 8 inclusive
		}
		else if (methCalcType == 2) { // exponential
			methLevel = 8 - getSummedMethLevel();
			if (methLevel > 0)
				methLevel = (2 << (methLevel-1)) - 1;
			rNum = Rndm::getRandomInt(prob->pr[ProbPistonBindToLigandPRI]); // range: 0 to 256 inclusive
		}
		else if (methCalcType == 3) { // sigmoidal
			methLevel = (8 - getSummedMethLevel()) - 4; // range: -4 to 4 inclusive
			methLevel = (int)((1.0f / (1.0f + exp((double)-methLevel))) * (double)ProbPistonBindToLigandPRI) - 4.0f;
			rNum = Rndm::getRandomInt(prob->pr[ProbPistonBindToLigandPRI]); // range: 0 to 256 inclusive
		}
		if (methLevel >= rNum) { // allow chance for fully methed to be bound to
		//rNum = Rndm::getRandomInt( ProbabilityMax );
		//if (rNum < prob->pr[ProbPistonBindToLigandPRI]) {
			lig = grid->getPolymer(ligID);
			lig->setAttached( true );
			for (col = rCol; col < rCol+len; col++) {
				bindings[rRow*numCols + col] = ligID;
				// also set binding for that ligand
				// get grid coordinates for monomer in this polymer
				shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
				// get coordinates of corresponding monomer in other polymer
				lig->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
				lig->setBindingVal( opRow, opCol, id );
			}

			if (ruleNum == ActivatePistonBindingRUL)
				pistonOffset[rRow] = 1;
			else // InhibitPistonBindingRUL
				pistonOffset[rRow] = 0;
			if (Msg::appM)
				printf( "doRule_PistonBinding\n" );
		}
	}
}

////////////////////////////////
// doRule_ProteinProteinBinding
// Protein-Protein Binding Rule
////////////////////////////////
void Protein::doRule_ProteinProteinBinding( enum ruleEnum ruleNum, int rRow, int rCol )
// Try to match CP protein with two TMP proteins
// assume this is a CP protein
// and assume that binding site is 3 or 4 or 5 monomers long
{
	if (attached) return;
	//printf( "0" );
	int col = 0;
	int count = 0;
	int gridPos = 0;
	enum MonomerTypeEnum contents = NullMN;
	gint* mGrid = grid->getLayerMono( Protein1LY );
	gint* mPolyGrid = grid->getLayerPoly( Protein1LY );
	int opID = 0; // other protein
	Polymer *op = NULL;
	// It almost certainly doesn't make any difference whether maxLen is set
	// here to 5 or 6 in the bacterial chemotaxis app. I'm just being extra careful.
#ifdef LOGIC_GATES_APP
	int maxLen = 6; // maximum length of binding region
#else
	int maxLen = 5; // maximum length of binding region
#endif
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer

	if (maxLen > numCols-rCol) maxLen = numCols-rCol;
	for (col = rCol; col < rCol+maxLen; col++) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		contents = (enum MonomerTypeEnum)(shape[rRow * numCols + col]);
		if ((contents == TurnMN) || (contents == WaterMN))
			break;
		if (contents == DelimitMN)
			break;
		if (bindings[rRow*numCols + col] != NullBNG) { // can't bind if already bound to something
			return;
		}
		if (matchP_P( contents, (MonomerTypeEnum)mGrid[gridPos] )) {
			count++;
		}
		else {
			break;
		}
		// if the monomer matches water, then the water won't have a valid opID
		//   only used in Logic Types application; allows protein to overhang over water
		if (contents != MatchWaterMN) {
			opID = mPolyGrid[gridPos];
			op = grid->getPolymer(opID);
			op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
			if (opBind != NullBNG) { // can't bind if other protein is already bound to something at that postion
				//printf( "Other is already bound\n" );
				return;
			}
		}
	}
	if ((count == (numCols - rCol)) // changed July 31
		|| (contents == DelimitMN) || (contents == TurnMN)
		|| (contents == WaterMN)) { // matched a protein
		if (op->getAttached()) { // only bind if other polymer is attached to something
			setAttached( true );

			// set bindings for both proteins
			// or it may have bound to multiple adjacent proteins
			for (col = rCol; col < rCol+maxLen; col++) {
				shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
				gridPos = gRow * gridSizeX + gCol;
				contents = (enum MonomerTypeEnum)(shape[rRow * numCols + col]);
				if ((contents == DelimitMN) || (contents == TurnMN) || (contents == WaterMN))
					break;
				opID = mPolyGrid[gridPos];
				if (opID == NullPolyID) { // nothing in grid at that position; fixes bug?
					if (Msg::errorM)
						printf( "ERROR: doRule_ProteinProteinBinding() bad opID=%d\n", opID );
					break;
				}

				op = grid->getPolymer(opID);
				op->setAttached( true );
				bindings[rRow*numCols + col] = opID;
				op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
				op->setBindingVal( opRow, opCol, id );
			}
#ifdef LOGIC_GATES_APP
			printf( "%c", 0x07 ); // beep
#endif
			if (Msg::appM)
				printf( "doRule_ProteinProteinBinding Matched: %s to %s\n",
					this->name, op->getName() );
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// doRule_ProteinMembraneBinding
// Protein binding to a membrane; especially trans-membrane protein (TMP)
//////////////////////////////////////////////////////////////////////////
void Protein::doRule_ProteinMembraneBinding( enum ruleEnum ruleNum, int rRow, int rCol )
{
	if (attached) return;
	int gridPos = 0;
	enum MonomerTypeEnum contents = NullMN;
	gint* mGrid = grid->getLayerMono( MembraneLY );
	gint* mPolyGrid = grid->getLayerPoly( MembraneLY );
	int opID = 0; // id of lipid bilayer
	Polymer *op = NULL;
	int pOff = getPistonOffset(rRow);

	// must match lipid bilayer exactly; Water Lipid Lipid Water
	shape2grid( GposContS2G, rRow, rCol-pOff, &gridPos, (int*)&contents );
	if (mGrid[gridPos] != WaterMN) return;
	shape2grid( GposContS2G, rRow, rCol+1-pOff, &gridPos, (int*)&contents );
	if (mGrid[gridPos] != Lip1MN) return;
	shape2grid( GposContS2G, rRow, rCol+2-pOff, &gridPos, (int*)&contents );
	if (mGrid[gridPos] != Lip1MN) return;
	opID = mPolyGrid[gridPos]; // get id of lipid bilayer while it's available
	if (opID != container->getID()) return; // only allowed to embed protein in outer membrane
	shape2grid( GposContS2G, rRow, rCol+3-pOff, &gridPos, (int*)&contents );
	if (mGrid[gridPos] != WaterMN) return;
	if (Msg::infoM) printf( "doRule_ProteinMembraneBinding()\n" );

	int col;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer
	op = grid->getPolymer(opID);
	setAttached( true );
	op->setAttached( true );

	// set bindings for both (only the two lipid overlaps)
	for (col = rCol+1; col < rCol+3; col++) {
		bindings[rRow*numCols + col] = opID;
		// also set binding for that ligand
		// get grid coordinates for monomer in this polymer
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		// get coordinates of corresponding monomer in other polymer
		op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
		op->setBindingVal( opRow, opCol, id );
	}
	if (Msg::appM)
		printf( "doRule_ProteinMembraneBinding\n" );
}

//////////////////////////////////////////////////////////////////////////////////////
// doRule_PhosphorylationCheA3
// Phosphorylation rule
// If monomer in shape at this one position matches grid (=D)
//   If monomer at postion +4 is phosphorylated
//     If something is bound to domain 2
//       move phosphoryl group from +4 to protein attached to domain 2
// CheA phosphorylation of CheY if MCP piston is extended and if CheA itself has phos
//////////////////////////////////////////////////////////////////////////////////////
void Protein::doRule_PhosphorylationCheA3( enum ruleEnum ruleNum, int rRow, int rCol )
{
	if (!attached) return;
	int gridPos = 0;
	enum MonomerTypeEnum contents = NullMN;
	gint* mGrid = grid->getLayerMono( Protein1LY );
	gint* mPolyGrid = grid->getLayerPoly( Protein1LY );
	int opID = 0; // id of lipid bilayer
	Protein *op = NULL;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer
	int opPos;

	shape2grid( GposContS2G, rRow, rCol, &gridPos, (int*)&contents );
	if (contents != mGrid[gridPos]) { // eColi way
		if (isPhosphorylated( rRow, rCol+4 )) {
			opPos = (rRow-1)*numCols + rCol+1; // expected position of other protein
			opID = bindings[opPos];
			if (opID != NullBNG) { // if it's bound
				dephosphorylate( rRow, rCol+4);
				op = (Protein*)grid->getPolymer(opID);
				shape2grid( GyGxS2G, rRow-1, rCol+1, &gRow, &gCol );
				op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
				op->phosphorylate( opRow, opCol );
				op->setChanged( true );
				if (Msg::appM)
					printf( "doRule_PhosphorylationCheA3\n" );
			}
		}
	}
}

///////////////////////////////////////////
// doRule_SwapOnPhosChange
// swap rules at binding site in a protein
///////////////////////////////////////////
void Protein::doRule_SwapOnPhosChange( enum ruleEnum ruleNum, int rRow, int rCol )
{
	if (ruleNum == SwapIfPhosphorylatedRUL) {
		if (isPhosphorylated( rRow-1, rCol-1 )) {
			setRuleVal( rRow, rCol, SwapIfUnphosphorylatedRUL );
		}
		else return;
	}
	else { // SwapIfUnphosphorylatedRUL
		if (!isPhosphorylated( rRow-1, rCol-1 )) {
			setRuleVal( rRow, rCol, SwapIfPhosphorylatedRUL );
		}
		else return;
	}
	if (Msg::appM)
		printf( "doRule_SwapOnPhosChange %d\n", ruleNum );
	int temp = getShapeVal( rRow, rCol );
	setShapeVal( rRow, rCol, getShapeVal( rRow, rCol+1 ) );
	setShapeVal( rRow, rCol+1, temp );
}

//////////////////////////////
// doRule_AutoPhosphorylation
// CheA
//////////////////////////////
void Protein::doRule_AutoPhosphorylation( enum ruleEnum ruleNum, int rRow, int rCol )
{
	if (isPhosphorylated( rRow, rCol )) return; // if is already phosphorylated

	int gridPos = 0;
	enum MonomerTypeEnum contents = NullMN;
	gint* mGrid = grid->getLayerMono( SmallMolLY );
	gint* mPolyGrid = grid->getLayerPoly( SmallMolLY );
	int axpID = 0; // Atp/Adp/Amp/Adenosine
	Polymer *axp = NULL;
	int mcpID = 0; // methyl-accepting chemotaxis proteins (Tar etc.)
	Protein *mcp;
	int col;
	int rNum = -1; // random number; initially < methLevel
	int methLevel = 0;
	int methCalcType = 2; // 1=linear, 2=exponential, 3=sigmoidal

	shape2grid( GposContS2G, rRow, rCol, &gridPos, (int*)&contents );
	if (mGrid[gridPos] == PhosphorylGroupMN) {
		axpID = mPolyGrid[gridPos];
		axp = grid->getPolymer( axpID );
		// search for last phosphoryl group in ATP/ADP/AMP
		for (col = axp->getNumCols()-1; col > 0; col--) {
			if (axp->getShapeVal( 0, col) == PhosphorylGroupMN) {

				// if this is a CheA,
				//   then its action depends on methylation level in Tar,etc.
				if (getRuleVal( rRow, rCol-4 ) == PhosphorylationCheA3RUL) {
					mcpID = getBindingVal( rRow+1, rCol-4 ); // DRKR
					if (mcpID != NullBNG) {
						mcp = (Protein*)grid->getPolymer( mcpID );
						if (methCalcType == 1) { // linear
							methLevel = mcp->getSummedMethLevel(); // range: 0 to 8 inclusive
							rNum = Rndm::getRandomInt(prob->pr[ProbCheAAutoPhosPRI]); // range: 0 to 7 inclusive
						}
						else if (methCalcType == 2) { // exponential
							methLevel = mcp->getSummedMethLevel();
							if (methLevel > 0)
								methLevel = (2 << (methLevel-1)) - 1;
							rNum = Rndm::getRandomInt(prob->pr[ProbCheAAutoPhosPRI]); // 255 ); // range: 0 to 255 inclusive
						}
						else if (methCalcType == 3) { // sigmoidal
							methLevel = mcp->getSummedMethLevel() - 4; // range: -4 to 4 inclusive
							methLevel = (int)((1.0f / (1.0f + exp((double)-methLevel))) * 255.0f) - 4; // range: 0 to 238
							rNum = Rndm::getRandomInt(prob->pr[ProbCheAAutoPhosPRI]); // 255 ); // range: 0 to 7 inclusive
						}
					}
				}

				if (methLevel > rNum) {
					axp->setShapeVal( 0, col, WaterMN ); // remove from ATP/ADP/AMP
					phosphorylate( rRow, rCol );
					if (Msg::appM)
						printf( "doRule_AutoPhosphorylation\n" );
					break; // only do max of one transfer
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////
// getSummedMethLevel
// Calculate a summation of the current methylation level
// for now this is just a straight arithmetic summation
// but it probably should be a more complex function
//////////////////////////////////////////////////////////
int Protein::getSummedMethLevel( void )
{
	int col = numCols-1;
	int row = 0; // piston row
	int sum = 0;

	// move up the row looking for first possibly methylated monomer (Glu E)
	while (col >= 0) {
		if (( getShapeVal( row, col ) == GluMN )
			|| isMethylated( row, col )) {
			sum += getMethLevel( row, col );
			break;
		}
		col--;
	}
	col--;

	// look for additional adjacent possibly methylated monomers
	while (col >= 0) {
		if (( getShapeVal( row, col ) == GluMN )
			|| isMethylated( row, col )) {
			sum += getMethLevel( row, col );
			col--;
		}
		else
			break;
	}
	return sum;
}

////////////////
// getMethLevel
////////////////
int Protein::getMethLevel( int mRow, int mCol )
{
	int pos = mRow*numCols + mCol;
	int methLevel = 0;
	switch( shape[pos] ) {
	case GluMN: // E
		methLevel = 0;
		break;
	case GluMeth1MN:
		methLevel = 1;
		break;
	case GluMeth2MN:
		methLevel = 2;
		break;
	case GluMeth3MN:
		methLevel = 3;
		break;
	case GluMeth4MN:
		methLevel = 4;
		break;
	default:
		methLevel = 0;
		break;
	}
	return methLevel;
}

////////////////////////////////
// doRule_AutoDephosphorylation
////////////////////////////////
void Protein::doRule_AutoDephosphorylation( enum ruleEnum ruleNum, int rRow, int rCol )
{
	if (!isPhosphorylated( rRow, rCol )) return; // if not already phosphorylated

	int gridPos = 0;
	enum MonomerTypeEnum contents = NullMN;
	gint* mGrid = grid->getLayerMono( SmallMolLY );
	gint* mPolyGrid = grid->getLayerPoly( SmallMolLY );
	int axpID = 0;
	Polymer *axp = NULL;
	int axpCol;
	int rNum;
	int probab = 1;

	shape2grid( GposContS2G, rRow, rCol, &gridPos, (int*)&contents );
	if ((mGrid[gridPos] == PhosphorylGroupMN) || (mGrid[gridPos] == AdenosineMN)) {
		// found a potential phosphoryl acceptor
		axpID = mPolyGrid[gridPos];
		axp = grid->getPolymer( axpID );
		// search for first non-phosphoryl group in ATP/ADP/AMP
		for (axpCol = 0; axpCol < axp->getNumCols(); axpCol++) {
			if (axp->getShapeVal( 0, axpCol) == WaterMN) {
				// determine probability based on whether I'm a CheY or CheB
				switch (getShapeVal( rRow+1, 0 )) {
				case LysMN: // CheY K
					probab = prob->pr[ProbCheYAutoDephosPRI];
					break;
				case AspMN: // CheB D
					probab = prob->pr[ProbCheBAutoDephosPRI];
					break;
				default:
					printf( "ERROR: doRule_AutoDephosphorylation default %d\n",
						getShapeVal( rRow+1, 0 ) );
					break;
				}
				rNum = Rndm::getRandomInt( ProbabilityMax-1 );
				if (rNum < probab) {
					axp->setShapeVal( 0, axpCol, PhosphorylGroupMN ); // add to ATP/ADP/AMP
					dephosphorylate( rRow, rCol );
					setChanged( true );
					if (Msg::appM)
						printf( "doRule_AutoDephosphorylation\n" );
				}
				break; // only do max of one transfer
			}
		}
	}
}

////////////////
// doRule_Motor
// Motor
////////////////
void Protein::doRule_Motor( enum ruleEnum ruleNum, int rRow, int rCol )
{
	int rNum; // random integer
	int row;
	// changed to 500 after increased size to 26 Tar
	int biasCalibrationTerm = 500; //550;
	LipidBilayer *eColi = (LipidBilayer*)container;

	// Calculate biased direction of movement
	// assume every row can have a CheY binding domain
	int bias = 0; // number of cheY currently binding to this motor
	for (row = 0; row < numRows; row++) {
		if (getBindingVal( row, rCol ) != NullBNG)
			bias += biasCalibrationTerm; // not sure what this value should be
	}

	// Calculate unbiased direction of movement; default when no CheY are attached
	rNum = Rndm::getRandomInt( ProbabilityMax-1 );

	// Sum or otherwise combine the unbiased and biased terms
	rNum -= bias;

	// 92% of time each individual motor should turn CCW
	if (rNum >= prob->pr[ProbMotorCwPRI]) {
		eColi->incCurrentCCW();
	}

	// motors always turn; only the direction changes
	// whether it's flipped or not is just a way of visually showing motor rotation
	if (flipped == FlipNone)
		flipped = FlipVert;
	else
		flipped = FlipNone;
}

//////////////////////////////////////////////////////////////////////////////////////
// doRule_MethylTransferase
// CheR
// Shimizu (2003) p.307 states: "methyltransferase CheR binds only inactive receptor"
//   - I assume this means when the piston is not extended
//////////////////////////////////////////////////////////////////////////////////////
void Protein::doRule_MethylTransferase( enum ruleEnum ruleNum, int rRow, int rCol )
{
	int col = 0;
	int gridPos = 0;
	int contents = 0;
	int len = 4; // number of monomers to match
	int rNum; // random integer

	if (len > numCols) return;

	gint* mGrid = grid->getLayerMono( Protein1LY );
	gint* mPolyGrid = grid->getLayerPoly( Protein1LY );
	gint* mGridSM = grid->getLayerMono( SmallMolLY );
	gint* mPolyGridSM = grid->getLayerPoly( SmallMolLY );
	int opID = 0; // other protein
	Protein *op = NULL;
	int methID = 0; // Methanol
	SmallMolecule *meth = NULL;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer

	// does CheR domain match domain in the grid (in other protein)
	for (col = rCol; col < rCol+len; col++) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		contents = getShapeVal( rRow, col );
		if (matchP_P( (MonomerTypeEnum)contents, (MonomerTypeEnum)mGrid[gridPos] ))
			continue;
		if ((contents == GluMN)
			&& (mGrid[gridPos]>=GluMeth1MN && mGrid[gridPos]<=GluMeth4MN))
			continue;
		return;
	}

	// is there a Methanol SM at the same location in the grid?
	len = 2; // only the first 2 monomers can be methylated
	for (col = rCol; col < rCol+len; col++) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		if ((mGridSM[gridPos] == MethylGroupMN) || (mGridSM[gridPos] == MethBaseMN)) {
			if ((mGrid[gridPos]==GluMN)
				|| (mGrid[gridPos]>=GluMeth1MN && mGrid[gridPos]<=GluMeth4MN)) {
				opID = mPolyGrid[gridPos];
				op = (Protein*)grid->getPolymer( opID );
				op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
				// The Shimizu decision should be based on the contents of opRow+1,
				// rather than on the piston offset which is a protein state.
				//if (op->getPistonOffset(0) == 1) { // Shimizu requirement
				if (op->getPistonOffset(0) == 0) { // inverse of Shimizu requirement
					rNum = Rndm::getRandomInt( ProbabilityMax );
					if (rNum > prob->pr[ProbCheRmethActivePRI] )
						return;
				}
				methID = mPolyGridSM[gridPos];
				meth = (SmallMolecule*)grid->getPolymer( methID );
				// search for first non-phosphoryl group in Methanol
				for (int mcol = meth->getNumCols()-1; mcol > 0; mcol--) {
					if (meth->getShapeVal( 0, mcol) == MethylGroupMN) {
						meth->setShapeVal( 0, mcol, WaterMN ); // 
						op->methylate( opRow, opCol );
						if (Msg::appM)
							printf( "doRule_MethylTransferase\n" );
						break; // only do max of one transfer
					}
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////
// doRule_MethylEsterase
// CheB
// an esterase is an enzyme so it needs to work with two other polymers
// - a protein that is methylated
// - a methylGroup acceptor
// according to Alon et al (1999) p.170:
//   - "feedback in which CheB demethylates receptors only when they are in their
//      active conformation" as the "molecular mechanism that leads to robust
//      exact adaptation"
//   - I assume this is when piston is extended ?
// Shimizu (2003) p.307 agrees: "methylesterase CheB binds only active receptor"
// But I'm allowing some probabilistic relaxation of Alon/Shimizu requirement
//   - to allow for influence between MCPs.
//////////////////////////////////////////////////////////////////////////////////
void Protein::doRule_MethylEsterase( enum ruleEnum ruleNum, int rRow, int rCol )
{
	int col = 0;
	int gridPos = 0;
	int contents = 0;
	int len = 4; // number of monomers to match
	int rNum; // random integer

	if (len > numCols) return;

	gint* mGrid = grid->getLayerMono( Protein1LY );
	gint* mPolyGrid = grid->getLayerPoly( Protein1LY );
	gint* mGridSM = grid->getLayerMono( SmallMolLY );
	gint* mPolyGridSM = grid->getLayerPoly( SmallMolLY );
	int opID = 0; // other protein
	Protein *op = NULL;
	int methID = 0; // Methanol
	SmallMolecule *meth = NULL;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer

	// does CheB domain match domain in the grid (in other protein)
	for (col = rCol; col > rCol-len; col--) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		contents = getShapeVal( rRow, col );
		if (matchP_P( (MonomerTypeEnum)contents, (MonomerTypeEnum)mGrid[gridPos] ))
			continue;
		if ((contents == GluMN)
			&& (mGrid[gridPos]>=GluMeth1MN && mGrid[gridPos]<=GluMeth4MN))
			continue;
		return;
	}

	// is there a Methanol SM at the same location in the grid?
	len = 2; // only the first 2 monomers can be methylated
	for (col = rCol; col > rCol-len; col--) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		if ((mGridSM[gridPos] == MethylGroupMN) || (mGridSM[gridPos] == MethBaseMN)) {
			if (mGrid[gridPos]>=GluMeth1MN && mGrid[gridPos]<=GluMeth4MN) {
				// found a potential methyl acceptor
				opID = mPolyGrid[gridPos];
				op = (Protein*)grid->getPolymer( opID );
				// The Alon decision should be based on the contents of opRow+1,
				// rather than on the piston offset which is a protein state.
				//if (op->getPistonOffset(0) == 0) { // Alon requirement
				if (op->getPistonOffset(0) == 1) { // inverse of Alon requirement
					rNum = Rndm::getRandomInt( ProbabilityMax );
					if (rNum > prob->pr[ProbCheBdemethInactivePRI] ) {
						//printf( "NO %d %d %d ", prob->pr[ProbCheBdemethInactive], rNum, ProbabilityMax );
						return;
					}
				}
				op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
				methID = mPolyGridSM[gridPos];
				meth = (SmallMolecule*)grid->getPolymer( methID );
				// search for first non-phosphoryl group in Methanol
				for (int mcol = 0; mcol < meth->getNumCols(); mcol++) {
					if (meth->getShapeVal( 0, mcol) == WaterMN) {
						meth->setShapeVal( 0, mcol, MethylGroupMN ); // add to Methanol
						op->demethylate( opRow, opCol );
						op->setChanged( true ); // is this needed ?
						if (Msg::appM)
							printf( "doRule_MethylEsterase\n" );
						break; // only do max of one transfer
					}
				}
			}
		}
	}
}

//////////////////////
// doRule_Phosphatase
// CheZ
//////////////////////
void Protein::doRule_Phosphatase( enum ruleEnum ruleNum, int rRow, int rCol )
{
	int col = 0;
	int gridPos = 0;
	int contents = 0;
	int len = 4; // number of monomers to match

	if (len > numCols) return;

	gint* mGrid = grid->getLayerMono( Protein2LY );
	gint* mPolyGrid = grid->getLayerPoly( Protein2LY );
	gint* mGridSM = grid->getLayerMono( SmallMolLY );
	gint* mPolyGridSM = grid->getLayerPoly( SmallMolLY );
	int opID = 0; // other protein
	Protein *op = NULL;
	int axpID = 0; // ATP etc.
	SmallMolecule *axp = NULL;
	int gRow, gCol; // grid row & col
	int opRow, opCol, opCont, opBind; // other polymer

	// does CheZ domain match domain in the grid (in other protein)
	for (col = rCol; col < rCol+len; col++) {
		shape2grid( GyGxS2G, rRow, col, &gRow, &gCol );
		gridPos = gRow * gridSizeX + gCol;
		contents = getShapeVal( rRow, col );
		if (matchP_P( (MonomerTypeEnum)contents, (MonomerTypeEnum)mGrid[gridPos] ))
			continue;
		return;
	}
	// have matched either a CheY~P or a motor or another CheZ

	// is there a Axp SM at the same location in the grid?
	shape2grid( GyGxS2G, rRow, rCol, &gRow, &gCol );
	gridPos = gRow * gridSizeX + gCol;
	if ((mGridSM[gridPos] == PhosphorylGroupMN) || (mGridSM[gridPos] == AdenosineMN)) {
		// found a potential methyl acceptor
		opID = mPolyGrid[gridPos];
		op = (Protein*)grid->getPolymer( opID );
		if (op->getNumRows() == numRows) return; // probably another CheZ
		op->grid2shape( 0, gRow, gCol, &opRow, &opCol, &opCont, &opBind );
		// check that the row and col values will not go out of range for motors
		if ((opRow < 1) || (opCol < 3)) return;
		// if it's CheY~P (and not motor) should be phosphorylated there
		if ( !op->isPhosphorylated( opRow-1, opCol-3 )) return;
		// It's a CheY~P
		// CheY~P must be free and not bound to motor; monomer can only bind 1 thing at a time
		if ( opBind != NullBNG) return;
		axpID = mPolyGridSM[gridPos];
		axp = (SmallMolecule*)grid->getPolymer( axpID );
		// search for first non-phosphoryl group in AXP
		for (int acol = 0; acol < axp->getNumCols(); acol++) {
			if (axp->getShapeVal( 0, acol) == WaterMN) {
				axp->setShapeVal( 0, acol, PhosphorylGroupMN ); // add to AXP
				op->dephosphorylate( opRow-1, opCol-3 );
				op->setChanged( true ); // is this needed ?
				if (Msg::appM)
					printf( "doRule_Phosphatase %d %s\n", op->getID(), op->getName() );
				break; // only do max of one transfer
			}
		}
	}
}

/////////////////////////////////////////////////////////////////////
// isPhosphorylated
// Is the monomer at this row and col in the protein phosphorylated?
/////////////////////////////////////////////////////////////////////
bool Protein::isPhosphorylated( int row, int col )
{
	enum MonomerTypeEnum monomer = (enum MonomerTypeEnum)shape[row*numCols + col];
	if ((monomer >= HisPhosMN) && (monomer <= TyrPhosMN))
		return true;
	else
		return false;
}

/////////////////////////////////////////////////////////////////
// phosphorylate
// Phosphorylate the monomer at this row and col in the protein.
/////////////////////////////////////////////////////////////////
void Protein::phosphorylate( int row, int col )
{
	int pos = row*numCols + col;
	switch( shape[pos] ) {
	case HisMN:
		shape[pos] = HisPhosMN;
		break;
	case SerMN:
		shape[pos] = SerPhosMN;
		break;
	case ThrMN:
		shape[pos] = ThrPhosMN;
		break;
	case TyrMN:
		shape[pos] = TyrPhosMN;
		break;
	default:
		break;
	}
}

/////////////////////////////////////////////////////////////////
// dephosphorylate
// Dephosphorylate the monomer at this row and col in the protein.
/////////////////////////////////////////////////////////////////
void Protein::dephosphorylate( int row, int col )
{
	int pos = row*numCols + col;
	switch( shape[pos] ) {
	case HisPhosMN:
		shape[pos] = HisMN;
		break;
	case SerPhosMN:
		shape[pos] = SerMN;
		break;
	case ThrPhosMN:
		shape[pos] = ThrMN;
		break;
	case TyrPhosMN:
		shape[pos] = TyrMN;
		break;
	default:
		break;
	}
}

////////////////////////////////////////////////////////////////
// isMethylated
// Is the monomer at this row and col in the protein methylated?
/////////////////////////////////////////////////////////////////
bool Protein::isMethylated( int row, int col )
{
	enum MonomerTypeEnum monomer = (enum MonomerTypeEnum)shape[row*numCols + col];
	if ((monomer >= GluMeth1MN) && (monomer <= GluMeth4MN))
		return true;
	else
		return false;
}

/////////////////////////////////////////////////////////////
// methylate
// Methylate the monomer at this row and col in the protein.
/////////////////////////////////////////////////////////////
void Protein::methylate( int row, int col )
{
	int pos = row*numCols + col;
	switch( shape[pos] ) {
	case GluMN:
		shape[pos] = GluMeth1MN;
		break;
	case GluMeth1MN:
		shape[pos] = GluMeth2MN;
		break;
	case GluMeth2MN:
		shape[pos] = GluMeth3MN;
		break;
	case GluMeth3MN:
		shape[pos] = GluMeth4MN;
		break;
	default:
		break;
	}
}

/////////////////////////////////////////////////////////////
// demethylate
// Demethylate the monomer at this row and col in the protein.
/////////////////////////////////////////////////////////////
void Protein::demethylate( int row, int col )
{
	int pos = row*numCols + col;
	switch( shape[pos] ) {
	case GluMeth4MN:
		shape[pos] = GluMeth3MN;
		break;
	case GluMeth3MN:
		shape[pos] = GluMeth2MN;
		break;
	case GluMeth2MN:
		shape[pos] = GluMeth1MN;
		break;
	case GluMeth1MN:
		shape[pos] = GluMN;
		break;
	default:
		break;
	}
}

//////////////////////////////////////////
// inventory
// Save out an inventory of all polymers.
//////////////////////////////////////////
void Protein::inventory( FILE *fp, Monomers *mn )
{
	int row, col;
	bool evenRow = true; // even row is row 0, 2, 4, etc.
	int mon; // monomer
	int bin; // binding
	int rul; // rule
	char brOut[100]; // bindings and rules info
	memset(brOut, 0, 100);
	int count = 0;
	if (getGhost())
		fprintf( fp, "Pg,%5d,", id );
	else
		fprintf( fp, "P,%5d,", id );
	// output the shape
	for (row = 0; row < numRows; row++) {
		for (col = 0; col < numCols; col++) {
			if (evenRow) {
				mon = getShapeVal( row, col );
				bin = getBindingVal( row, col );
				rul = getRuleVal( row, col );
			}
			else {
				mon = getShapeVal( row, numCols-col-1 );
				bin = getBindingVal( row, numCols-col-1 );
				rul = getRuleVal( row, numCols-col-1 );
			}
			if ((mon != NullMN) && (mon != WaterMN)) {
				fprintf( fp, "%c", mn->get(mon)->code );
				if (bin == NullBNG) {
					if (rul == NullRUL)
						brOut[count] = ' '; // no binding, no rule
					else {
						if (rul == NoMoreRulesRUL)
							brOut[count] = '.';
						else
							brOut[count] = '='; // no binding, has rule
					}
				}
				else { // has a binding
					if (rul == NullRUL)
						brOut[count] = '|'; // has binding, no rule
					else {
						if (rul == NoMoreRulesRUL)
							brOut[count] = '|';
						else
							brOut[count] = '#'; // has binding, has rule
					}
				}
				count++;
			}
		}
		if (evenRow) evenRow = false;
		else evenRow = true;
	}
	fprintf( fp, ",%s\n", getName() );
	fprintf( fp, " ,     ,%s, \n", brOut );
}

////////////////////////////
// doStats
// Save various statistics.
////////////////////////////
void Protein::doStats( void )
{
	switch( statsType ) {
	case Mcp_ST:
		grid->stats->cnt[Mcp_ST]++;
		if (pistonOffset[0] == 1)
			grid->stats->cnt[Mcp_PistonBound_ST]++;
		grid->stats->cnt[Mcp_Methylated_ST] += getSummedMethLevel();
		break;
	case CheW_ST:
		grid->stats->cnt[CheW_ST]++;
		break;
	case CheA_ST:
		grid->stats->cnt[CheA_ST]++;
		if (isPhosphorylated(1,5))
			grid->stats->cnt[CheA_Phosphorylated_ST]++;
		if (getBindingVal(0,2) != NullBNG)
			grid->stats->cnt[CheA_CheYbBound_ST]++;
		break;
	case CheY_ST:
		grid->stats->cnt[CheY_ST]++;
		if (isPhosphorylated(0,0))
			grid->stats->cnt[CheY_Phosphorylated_ST]++;
		if (getBindingVal(0,0) != NullBNG)
			grid->stats->cnt[CheY_CheABound_ST]++;
		if (getBindingVal(1,0) != NullBNG)
			grid->stats->cnt[CheY_MotorBound_ST]++;
		break;
	case Motor_ST:
		grid->stats->cnt[Motor_ST]++;
		if (getBindingVal(0,1) != NullBNG)
			grid->stats->cnt[Motor_CheYBound_ST]++;
		if (getBindingVal(1,1) != NullBNG)
			grid->stats->cnt[Motor_CheYBound_ST]++;
		if (getBindingVal(2,1) != NullBNG)
			grid->stats->cnt[Motor_CheYBound_ST]++;
		break;
	case CheB_ST:
		grid->stats->cnt[CheB_ST]++;
		if (isPhosphorylated(0,0))
			grid->stats->cnt[CheB_Phosphorylated_ST]++;
		if (getBindingVal(0,0) != NullBNG)
			grid->stats->cnt[CheB_CheABound_ST]++;
		break;
	case CheR_ST:
		grid->stats->cnt[CheR_ST]++;
		break;
	case CheZ_ST:
		grid->stats->cnt[CheZ_ST]++;
		break;
	default:
		if (Msg::errorM)
			printf( "ERROR protein::doStats() default=[%d %s]", statsType, name );
		break;
	}
}
